import os
from .base import BaseData
from ..default import DATASET
from ..utils import VarDim, get_logger, check_dir_or_make
import torch
from time import time
from copy import copy
import pandas as pd
import numpy as np
import random
from os.path import dirname, join
import scipy.sparse as sp
from numbers import Number
from scipy.sparse import csr_matrix
from collections import defaultdict


def shuffle(*arrays, **kwargs):
    require_indices = kwargs.get('indices', False)

    if len(set(len(x) for x in arrays)) != 1:
        raise ValueError('All inputs to shuffle must have ' 'the same length.')

    shuffle_indices = np.arange(len(arrays[0]))
    np.random.shuffle(shuffle_indices)

    if len(arrays) == 1:
        result = arrays[0][shuffle_indices]
    else:
        result = tuple(x[shuffle_indices] for x in arrays)

    if require_indices:
        return result, shuffle_indices
    else:
        return result


def minibatch(*tensors, **kwargs):
    batch_size = kwargs['batch_size']

    if len(tensors) == 1:
        tensor = tensors[0]
        for i in range(0, len(tensor), batch_size):
            yield tensor[i : i + batch_size]
    else:
        for i in range(0, len(tensors[0]), batch_size):
            yield tuple(x[i : i + batch_size] for x in tensors)


def pairwise_sample(dataset):
    """
    the original impliment of BPR Sampling in LightGCN
    :return:
        np.array
    """
    user_num = dataset.traindataSize
    users = np.random.randint(0, dataset.n_users, user_num)
    allPos = dataset.allPos
    S = []

    for i, user in enumerate(users):
        posForUser = allPos[user]
        if len(posForUser) == 0:
            continue
        posindex = np.random.randint(0, len(posForUser))
        positem = posForUser[posindex]
        while True:
            negitem = np.random.randint(0, dataset.n_items)
            if negitem in posForUser:
                continue
            else:
                break
        S.append([user, positem, negitem])
    return np.array(S)


def pointwise_sample(dataset: BaseData, negative_num_each_interaction):
    data = []
    train_dict = dataset.info_describe()['train_dict']
    n_items = dataset.info_describe()['n_items']

    full_items = set(range(n_items))
    for uid, iids in train_dict.items():
        data.extend([(uid, iid, 1) for iid in iids])

        left_set = list(full_items - set(iids))
        negs = np.random.choice(
            left_set, size=len(iids) * negative_num_each_interaction
        )
        data.extend([(uid, ni, 0) for ni in negs])
    return np.array(data)


def csv2dict(file, filter=4):
    df = pd.read_csv(file)
    interactions = {}
    df = df.sort_values("timestamp")
    df = df[df['rating'] >= filter]
    for u, i in zip(df['user_id'], df['item_id']):
        if u not in interactions:
            interactions[int(u)] = []
        if i not in interactions[u]:  # to keep order of time
            interactions[int(u)].append(int(i))
    return interactions


def fake_array2dict(fake_array, n_users, filter_num=4):
    assert len(fake_array.shape) == 2, "Expect a user-item 2D rating matrix"
    uids, iids = np.where(fake_array > filter_num)
    uids += n_users
    result = defaultdict(list)
    for u, i in zip(uids, iids):
        result[u].append(i)
    return result


def npy2dict(file):
    return np.load(file, allow_pickle=True).item()


def convert2dict(file: str, filter_num):
    if file.endswith(".csv"):
        return csv2dict(file, filter=filter_num)
    elif file.endswith(".npy"):
        return npy2dict(file)
    else:
        raise ValueError(f"Expect data file ends witg [csv/npy], but got {file}")


class ImplicitData(BaseData):
    def __init__(self, **config):
        self.config = config
        self.logger = get_logger(
            f"{__name__}:{self.dataset_name}", level=config['logging_level']
        )
        self._mode = "train"
        self._load_data()
        self._init_data()

    def _load_data(self):
        train_file = self.config['path_train']
        valid_file = self.config['path_valid']
        test_file = self.config['path_test']

        for attr, default in zip(
            ['train_dict', 'valid_dict', 'test_dict'],
            [train_file, valid_file, test_file],
        ):
            if self.config[attr] is not None:
                self.logger.debug(f"Init {attr} from a passed dict")
                setattr(self, attr, self.config[attr])
                continue
            maybe_cache = join(
                self.config['cache_dir'],
                f"{self.dataset_name}_implicit_{attr}.npy",
            )
            if os.path.exists(maybe_cache) and self.config['if_cache']:
                self.logger.info(f"loading cached {attr}")
                setattr(self, attr, np.load(maybe_cache, allow_pickle=True).item())
            else:
                setattr(self, attr, convert2dict(default, self.config['rating_filter']))
                if self.config['if_cache']:
                    check_dir_or_make(self.config['cache_dir'])
                    np.save(maybe_cache, getattr(self, attr))

    def _init_data(self):
        self.n_users = 0
        self.n_items = 0

        self.split = self.config['A_split']
        self.folds = self.config['A_n_fold']

        (
            self.traindataSize,
            self.trainUniqueUsers,
            self.trainUser,
            self.trainItem,
        ) = self.read_data(self.train_dict)

        (
            self.validDataSize,
            self.validUniqueUsers,
            self.validUser,
            self.validItem,
        ) = self.read_data(self.valid_dict)

        (
            self.testDataSize,
            self.testUniqueUsers,
            self.testUser,
            self.testItem,
        ) = self.read_data(self.test_dict)

        self.n_items += 1
        self.n_users += 1

        self.Graph = None
        self.logger.debug(f"{self.traindataSize} interactions for training")
        self.logger.debug(f"{self.validDataSize} interactions for validation")
        self.logger.debug(f"{self.testDataSize} interactions for testing")
        self.logger.debug(
            f"{self.dataset_name} Sparsity : {(self.traindataSize + self.validDataSize + self.testDataSize) / self.n_users / self.n_items}"
        )

        # (users,items), bipartite graph
        self.UserItemNet = csr_matrix(
            (np.ones(len(self.trainUser)), (self.trainUser, self.trainItem)),
            shape=(self.n_users, self.n_items),
        )
        self.users_D = np.array(self.UserItemNet.sum(axis=1)).squeeze()
        self.users_D[self.users_D == 0.0] = 1
        self.items_D = np.array(self.UserItemNet.sum(axis=0)).squeeze()
        self.items_D[self.items_D == 0.0] = 1.0
        # pre-calculate
        self._allPos = self.getUserPosItems(list(range(self.n_users)))
        self.logger.debug(f"{self.dataset_name} is ready to go")

        if self.config['need_graph']:
            self.getSparseGraph()

    def read_data(self, data_dict):
        trainUniqueUsers, trainItem, trainUser = [], [], []
        traindataSize = 0
        for uid in data_dict.keys():
            if len(data_dict[uid]) != 0:
                trainUniqueUsers.append(uid)
                trainUser.extend([uid] * len(data_dict[uid]))
                trainItem.extend(data_dict[uid])
                self.n_items = max(self.n_items, max(data_dict[uid]))
                self.n_users = max(self.n_users, uid)
                traindataSize += len(data_dict[uid])

        self.trainUniqueUsers = np.array(trainUniqueUsers)
        self.trainUser = np.array(trainUser)
        self.trainItem = np.array(trainItem)
        return (
            traindataSize,
            np.array(trainUniqueUsers),
            np.array(trainUser),
            np.array(trainItem),
        )

    def getSparseGraph(self):
        self.logger.debug("loading adjacency matrix")
        if self.Graph is None:
            maybe_cache = join(
                self.config['cache_dir'],
                f"adj_mat_{self.dataset_name}_{self.n_users}_{self.n_items}.npz",
            )
            if os.path.exists(maybe_cache) and self.config['if_cache']:
                pre_adj_mat = sp.load_npz(maybe_cache)
                self.logger.info(
                    f"successfully loaded adj_mat from {maybe_cache}, this could cause the inconsistency of the dataset"
                )
                norm_adj = pre_adj_mat
            else:
                self.logger.debug("generating adjacency matrix")
                s = time()
                adj_mat = sp.dok_matrix(
                    (self.n_users + self.n_items, self.n_users + self.n_items),
                    dtype=np.float32,
                )
                adj_mat = adj_mat.tolil()
                R = self.UserItemNet.tolil()
                adj_mat[: self.n_users, self.n_users :] = R
                adj_mat[self.n_users :, : self.n_users] = R.T
                adj_mat = adj_mat.todok()
                # adj_mat = adj_mat + sp.eye(adj_mat.shape[0])

                rowsum = np.array(adj_mat.sum(axis=1))
                d_inv = np.power(rowsum + 1e-14, -0.5).flatten()
                d_inv[np.isinf(d_inv)] = 0.0
                d_mat = sp.diags(d_inv)

                norm_adj = d_mat.dot(adj_mat)
                norm_adj = norm_adj.dot(d_mat)
                norm_adj = norm_adj.tocsr()
                end = time()
                self.logger.debug(f"costing {end - s}s, saved norm_mat...")
                if self.config['if_cache']:
                    self.logger.debug(f"cache adj mat to {self.config['cache_dir']}")
                    check_dir_or_make(self.config['cache_dir'])
                    sp.save_npz(
                        join(
                            self.config['cache_dir'],
                            f"adj_mat_{self.dataset_name}_{self.n_users}_{self.n_items}.npz",
                        ),
                        norm_adj,
                    )

            if self.split == True:
                self.Graph = self._split_A_hat(norm_adj)
                self.logger.debug("split matrix")
            else:
                self.Graph = self._convert_sp_mat_to_sp_tensor(norm_adj)
                self.Graph = self.Graph.coalesce().to(self.config['device'])
                self.logger.debug("don't split the matrix")
        return self.Graph

    @property
    def allPos(self):
        return self._allPos

    def _split_A_hat(self, A):
        A_fold = []
        fold_len = (self.n_users + self.n_items) // self.folds
        for i_fold in range(self.folds):
            start = i_fold * fold_len
            if i_fold == self.folds - 1:
                end = self.n_users + self.n_items
            else:
                end = (i_fold + 1) * fold_len
            A_fold.append(
                self._convert_sp_mat_to_sp_tensor(A[start:end])
                .coalesce()
                .to(self.config['device'])
            )
        return A_fold

    def _convert_sp_mat_to_sp_tensor(self, X):
        coo = X.tocoo().astype(np.float32)
        row = torch.Tensor(coo.row).long()
        col = torch.Tensor(coo.col).long()
        index = torch.stack([row, col])
        data = torch.FloatTensor(coo.data)
        return torch.sparse.FloatTensor(index, data, torch.Size(coo.shape))

    def getUserItemFeedback(self, users, items):
        """
        users:
            shape [-1]
        items:
            shape [-1]
        return:
            feedback [-1]
        """
        return np.array(self.UserItemNet[users, items]).astype('uint8').reshape((-1,))

    def getUserPosItems(self, users):
        posItems = []
        for user in users:
            posItems.append(self.UserItemNet[user].nonzero()[1])
        return posItems

    @classmethod
    def from_config(cls, name, **user_config):
        args = list(DATASET['implicit'][name])
        return super().from_config("implicit", name, args, user_config)

    def batch_describe(self):
        if self.config['sample'] == 'pairwise' and self.mode() == "train":
            return {
                "users": (
                    torch.int64,
                    (VarDim(max=self.config['pairwise_batch_size'], comment="batch")),
                ),
                "positive_items": (
                    torch.int64,
                    (VarDim(max=self.config['pairwise_batch_size'], comment="batch")),
                ),
                "negative_items": (
                    torch.int64,
                    (VarDim(max=self.config['pairwise_batch_size'], comment="batch")),
                ),
            }
        if self.config['sample'] == 'pointwise' and self.mode() == "train":
            return {
                "users": (
                    torch.int64,
                    (VarDim(max=self.config['pointwise_batch_size'], comment="batch")),
                ),
                "items": (
                    torch.int64,
                    (VarDim(max=self.config['pointwise_batch_size'], comment="batch")),
                ),
                "labels": (
                    torch.int64,
                    (VarDim(max=self.config['pointwise_batch_size'], comment="batch")),
                ),
            }
        elif self.mode() in ["validate", "test"]:
            return {
                "users": (
                    torch.int64,
                    (VarDim(max=self.config['test_batch_size'], comment='batch')),
                ),
                "positive_items": (
                    list,
                    (VarDim(max=self.config['test_batch_size'], comment='batch')),
                ),
                "ground_truth": (
                    list,
                    (VarDim(max=self.config['test_batch_size'], comment='batch')),
                ),
            }

    def info_describe(self):
        infos = {
            "n_users": self.n_users,
            "n_items": self.n_items,
            "train_interactions": self.traindataSize,
            "valid_interactions": self.validDataSize,
            "test_interactions": self.testDataSize,
            "train_dict": self.train_dict,
            "valid_dict": self.valid_dict,
            "test_dict": self.test_dict,
            "batch_describe": self.batch_describe(),
        }
        if self.config['need_graph']:
            infos['graph'] = self.Graph
        return infos

    def mode(self) -> str:
        return self._mode

    def generate_batch(self, **config):
        if self.config['sample'] == 'pairwise' and self.mode() == 'train':
            S = pairwise_sample(self)
            users = torch.Tensor(S[:, 0]).long()
            posItems = torch.Tensor(S[:, 1]).long()
            negItems = torch.Tensor(S[:, 2]).long()

            users = users.to(self.config['device'])
            posItems = posItems.to(self.config['device'])
            negItems = negItems.to(self.config['device'])
            users, posItems, negItems = shuffle(users, posItems, negItems)
            # total_batch = (
            #     len(users) + self.config['pairwise_batch_size'] - 1
            # ) // self.config['pairwise_batch_size']
            for batch_users, batch_pos, batch_neg in minibatch(
                users, posItems, negItems, batch_size=self.config['pairwise_batch_size']
            ):
                yield {
                    "users": batch_users.to(self.config['device']),
                    "positive_items": batch_pos.to(self.config['device']),
                    "negative_items": batch_neg.to(self.config['device']),
                }
        elif self.config['sample'] == 'pointwise' and self.mode() == 'train':
            S = pointwise_sample(self, self.config['negative_ratio'])
            users = torch.Tensor(S[:, 0]).long()
            items = torch.Tensor(S[:, 1]).long()
            labels = torch.Tensor(S[:, 2]).long()

            users = users.to(self.config['device'])
            items = items.to(self.config['device'])
            labels = labels.to(self.config['device'])
            users, items, labels = shuffle(users, items, labels)
            # total_batch = (
            #     len(users) + self.config['pairwise_batch_size'] - 1
            # ) // self.config['pairwise_batch_size']
            for batch_users, batch_items, batch_labels in minibatch(
                users, items, labels, batch_size=self.config['pointwise_batch_size']
            ):
                yield {
                    "users": batch_users,
                    "items": batch_items,
                    "labels": batch_labels,
                }
        elif self.mode() == "train":
            raise NotImplementedError("Not implemented yet")
        elif self.mode() in ["validate", "test"]:
            if self.mode() == 'validate':
                testDict = self.valid_dict
            else:
                testDict = self.test_dict
            users = list(testDict.keys())
            for batch_users in minibatch(
                users, batch_size=self.config['test_batch_size']
            ):
                all_pos = self.getUserPosItems(batch_users)
                ground_true = [testDict[u] for u in batch_users]
                yield {
                    "users": torch.Tensor(batch_users).long().to(self.config['device']),
                    "positive_items": all_pos,
                    "ground_truth": ground_true,
                }

    def switch_mode(self, mode):
        assert mode in ["train", "test", "validate"]
        self._mode = mode

    def inject_data(self, data_mode, data, **kwargs):
        if data_mode == 'explicit':
            new_train_dict = copy(self.train_dict)
            inject_dict = fake_array2dict(
                data, self.n_users, filter_num=kwargs['filter_num']
            )
            for k, v in inject_dict.items():
                assert (
                    k not in new_train_dict
                ), f"Injection to a exist user {k} is not allowed"
                new_train_dict[k] = v
            return self.reset(train_dict=new_train_dict, if_cache=False)
        raise NotImplementedError(f"Injection not supported in {data_mode} mode")

    def partial_sample(self, **kwargs) -> 'BaseData':
        # TODO return the sampled implicit feedback, but attacker may want ratings
        assert "user_ratio" in kwargs, "Expect to have [user_ratio]"
        user_ratio = kwargs['user_ratio']
        if abs(user_ratio - 1) < 1e-9:
            return self
        users = list(self.train_dict)
        random.shuffle(users)
        left_users = users[: int(len(users) * user_ratio)]
        new_dict = {u: self.train_dict[u] for u in left_users}
        return self.reset(train_dict=new_dict)



from .base import BaseAttacker
import math
import torch
from torch import nn
import numpy as np
from functools import partial
from ...default import MODEL
from ...utils import pick_optim, VarDim, filler_filter_mat


class Aush(BaseAttacker):
    def __init__(self, **config) -> None:
        super().__init__()

        self.selected_ids = config['selected_ids']
        self.attack_num = config['attack_num']
        self.filler_num = config['filler_num']
        self.dataset = config['dataset']
        self.n_items = config['dataset'].info_describe()['n_items']
        self.config = config
        self.device = config['device']
        self.ZR_ratio = config['ZR_ratio']
        self.train_data_array = config['dataset'].info_describe()['train_mat']
        self.build_network()

    def build_network(self):
        self.netG = AushGenerator(input_dim=self.n_items).to(self.device)
        self.G_optimizer = pick_optim(self.config['optim_g'])(
            self.netG.parameters(), lr=self.config['lr_g']
        )

        self.netD = AushDiscriminator(input_dim=self.n_items).to(self.device)
        self.D_optimizer = pick_optim(self.config['optim_d'])(
            self.netD.parameters(), lr=self.config['lr_d']
        )

    @classmethod
    def from_config(cls, **kwargs):
        args = list(MODEL['attacker']['aush'])
        user_args = "dataset"
        return super().from_config("aush", args, user_args, kwargs)

    def info_describe(self):
        return super().info_describe()

    def input_describe(self):
        return {"train_step": {"target_id_list": (list, VarDim())}}

    def output_describe(self):
        return {
            "train_step": {
                "d_losses": (float, []),
                "g_loss_rec_l": (float, []),
                "g_loss_shilling_l": (float, []),
                "g_loss_gan_l": (float, []),
            }
        }

    def sample_fillers(self, real_profiles, target_id_list):
        fillers = np.zeros_like(real_profiles)
        filler_pool = (
            set(range(self.n_items)) - set(self.selected_ids) - set(target_id_list)
        )
        filler_sampler = lambda x: np.random.choice(
            size=self.filler_num,
            replace=True,
            a=list(set(np.argwhere(x > 0).flatten()) & filler_pool),
        )

        sampled_cols = np.array(
            [filler_sampler(x) for x in real_profiles], dtype="int64"
        )

        sampled_rows = np.repeat(np.arange(real_profiles.shape[0]), self.filler_num)
        fillers[sampled_rows, sampled_cols.flatten()] = 1
        return fillers

    def train_step(self, **config):
        target_id_list = config['target_id_list']
        bce_loss = nn.BCELoss()
        mse_loss = nn.MSELoss()

        g_loss_rec_l = []
        g_loss_shilling_l = []
        g_loss_gan_l = []
        d_losses = []

        index_filter = partial(
            filler_filter_mat,
            selected_ids=self.selected_ids,
            filler_num=self.filler_num,
            target_id_list=target_id_list,
        )
        for idx, dp in enumerate(
            self.dataset.generate_batch(user_filter=index_filter, **config)
        ):
            batch_set_idx = dp['users']
            real_profiles = dp['users_mat']

            valid_labels = (
                torch.ones_like(batch_set_idx)
                .type(torch.float)
                .to(self.device)
                .reshape(len(batch_set_idx), 1)
            )
            fake_labels = (
                torch.zeros_like(batch_set_idx)
                .type(torch.float)
                .to(self.device)
                .reshape(len(batch_set_idx), 1)
            )
            fillers_mask = self.sample_fillers(
                real_profiles.cpu().numpy(), target_id_list
            )

            # selected
            selects_mask = np.zeros_like(fillers_mask)
            selects_mask[:, self.selected_ids] = 1.0
            # target
            target_patch = np.zeros_like(fillers_mask)
            target_patch[:, self.selected_ids] = 5.0
            # ZR_mask
            ZR_mask = (real_profiles.cpu().numpy() == 0) * selects_mask
            pools = np.argwhere(ZR_mask)
            np.random.shuffle(pools)
            pools = pools[: math.floor(len(pools) * (1 - self.ZR_ratio))]
            ZR_mask[pools[:, 0], pools[:, 1]] = 0

            fillers_mask = torch.tensor(fillers_mask).type(torch.float).to(self.device)
            selects_mask = torch.tensor(selects_mask).type(torch.float).to(self.device)
            target_patch = torch.tensor(target_patch).type(torch.float).to(self.device)
            ZR_mask = torch.tensor(ZR_mask).type(torch.float).to(self.device)

            input_template = torch.mul(real_profiles, fillers_mask)
            # ----------generate----------
            self.netG.eval()
            gen_output = self.netG(input_template)
            gen_output = gen_output.detach()
            # ---------mask--------
            selected_patch = torch.mul(gen_output, selects_mask)
            middle = torch.add(input_template, selected_patch)
            fake_profiles = torch.add(middle, target_patch)
            # --------Discriminator------
            # forward
            self.D_optimizer.zero_grad()
            self.netD.train()
            d_valid_labels = self.netD(real_profiles * (fillers_mask + selects_mask))
            d_fake_labels = self.netD(fake_profiles * (fillers_mask + selects_mask))
            # loss
            D_real_loss = bce_loss(d_valid_labels, valid_labels)
            D_fake_loss = bce_loss(d_fake_labels, fake_labels)
            d_loss = 0.5 * (D_real_loss + D_fake_loss)
            d_loss.backward()
            self.D_optimizer.step()
            self.netD.eval()
            d_losses.append(d_loss.item())

            # ---------train G-------
            self.netG.train()
            d_fake_labels = self.netD(fake_profiles * (fillers_mask + selects_mask))
            g_loss_gan = bce_loss(d_fake_labels, valid_labels)
            g_loss_shilling = mse_loss(fake_profiles * selects_mask, selects_mask * 5.0)
            g_loss_rec = mse_loss(
                fake_profiles * selects_mask * ZR_mask,
                selects_mask * input_template * ZR_mask,
            )
            g_loss = g_loss_gan + g_loss_rec + g_loss_shilling

            g_loss_rec_l.append(g_loss_rec.item())
            g_loss_shilling_l.append(g_loss_shilling.item())
            g_loss_gan_l.append(g_loss_gan.item())
            self.G_optimizer.zero_grad()
            g_loss.backward(retain_graph=True)
            self.G_optimizer.step()
        return (
            np.mean(d_losses),
            np.mean(g_loss_rec_l),
            np.mean(g_loss_shilling_l),
            np.mean(g_loss_gan_l),
        )

    def generate_fake(self, **kwargs):
        target_id_list = kwargs['target_id_list']
        mask_array = (self.train_data_array > 0).astype('float')
        mask_array[:, self.selected_ids + target_id_list] = 0
        available_idx = np.where(np.sum(mask_array, 1) >= self.filler_num)[0]
        available_idx = np.random.permutation(available_idx)
        idx = available_idx[np.random.randint(0, len(available_idx), self.attack_num)]
        idx = list(idx)

        real_profiles = self.train_data_array[idx, :]
        # sample fillers
        fillers_mask = self.sample_fillers(real_profiles, target_id_list)
        # selected
        selects_mask = np.zeros_like(fillers_mask)
        selects_mask[:, self.selected_ids] = 1.0
        # target
        target_patch = np.zeros_like(fillers_mask)
        target_patch[:, target_id_list] = 5.0

        # Generate
        real_profiles = torch.tensor(real_profiles).type(torch.float).to(self.device)
        fillers_mask = torch.tensor(fillers_mask).type(torch.float).to(self.device)
        selects_mask = torch.tensor(selects_mask).type(torch.float).to(self.device)
        target_patch = torch.tensor(target_patch).type(torch.float).to(self.device)
        input_template = torch.mul(real_profiles, fillers_mask)
        self.netG.eval()
        gen_output = self.netG(input_template)
        selected_patch = torch.mul(gen_output, selects_mask)
        middle = torch.add(input_template, selected_patch)
        fake_profiles = torch.add(middle, target_patch)
        fake_profiles = fake_profiles.detach().cpu().numpy()
        # fake_profiles = self.generator.predict([real_profiles, fillers_mask, selects_mask, target_patch])
        # selected patches
        selected_patches = fake_profiles[:, self.selected_ids]
        selected_patches = np.round(selected_patches)
        selected_patches[selected_patches > 5] = 5
        selected_patches[selected_patches < 1] = 1
        fake_profiles[:, self.selected_ids] = selected_patches

        return fake_profiles

    def forward(self):
        pass


class AushGenerator(nn.Module):
    def __init__(self, input_dim):
        super(AushGenerator, self).__init__()
        self.main = nn.Sequential(
            nn.Linear(input_dim, 128),
            nn.Sigmoid(),
            nn.Linear(128, input_dim),
            nn.Sigmoid(),
        )

    def forward(self, input):
        return self.main(input) * 5


class AushDiscriminator(nn.Module):
    def __init__(self, input_dim):
        super(AushDiscriminator, self).__init__()
        self.main = nn.Sequential(
            nn.Linear(input_dim, 150),
            nn.Sigmoid(),
            nn.Linear(150, 150),
            nn.Sigmoid(),
            nn.Linear(150, 150),
            nn.Sigmoid(),
            nn.Linear(150, 1),
            nn.Sigmoid(),
        )

    def forward(self, input):
        return self.main(input)



import torch
from abc import ABC, abstractmethod
from functools import partial
from copy import copy
from ..default import WORKFLOW
from ..utils import get_logger, parse_args, user_side_check, type_if_long
from pprint import pprint

logger = get_logger(__name__)


class BaseWorkflow(ABC):
    @classmethod
    @abstractmethod
    def from_config(cls, name, arg_string, user_args, user_config):
        assert name in WORKFLOW, f"{name} is not on the default workflows"
        if not user_side_check(user_args, user_config):
            raise TypeError(f"Expect for user arguments [{user_args}]")
        # come from default
        args = parse_args(arg_string)
        user_args = parse_args(user_args)
        try:
            default_config = {k: copy(WORKFLOW[name][k]) for k in args}
        except:
            error_message = f"{args} contain non-legal keys for {cls}!, expect in {list(WORKFLOW[name])}"
            raise KeyError(error_message)
        filter_user_config = {}
        for k, v in user_config.items():
            if k in args or k in user_args:
                filter_user_config[k] = v
            else:
                logger.debug(f"Unexpected key [{k}] for {cls}")
        default_config.update(filter_user_config)
        return cls(**default_config)

    @abstractmethod
    def info_describe(self):
        """return the types of datasets and models this workflow needs"""
        raise NotImplementedError

    @abstractmethod
    def execute(self):
        """return the types of datasets and models this workflow needs"""
        raise NotImplementedError

    def print_help(self, **kwargs) -> str:
        info_des = self.info_describe()
        assert isinstance(
            info_des, dict
        ), "Wrong return type of info_describe, expected Dict"
        for k, v in info_des.items():
            if k == 'batch_describe':
                continue
            v = type_if_long(v)
            info_des[k] = v
        pprint(info_des)



import torch
from .base import BaseAttacker
import numpy as np
import pandas as pd
from ...utils import user_side_check, get_logger, VarDim
from ...default import MODEL
from collections.abc import Iterable


class RandomAttack(BaseAttacker):
    def __init__(self, **config):
        super(RandomAttack, self).__init__()
        train_kv_array = config['dataset'].info_describe()['train_kvr']
        self.n_items = config['dataset'].info_describe()['n_items']

        self.global_mean = np.mean(train_kv_array[:, 2])
        self.global_std = np.std(train_kv_array[:, 2])
        self.attack_num = config['attack_num']
        self.filler_num = config['filler_num']
        self.logger = get_logger(__name__, level=config['logging_level'])

    @classmethod
    def from_config(cls, **kwargs):
        args = list(MODEL['attacker']['random'])
        user_args = "dataset"
        return super().from_config("random", args, user_args, kwargs)

    def train_step(self, users, positive_items, negative_items, **config):
        pass

    def forward(self, users, items):
        pass

    def generate_fake(self, target_id_list, **config):
        # target_id_list =[1551, 2510, 1167, 2362, 2233, 2801, 905, 1976, 3239, 3262]
        fake_profiles = np.zeros(shape=[self.attack_num, self.n_items], dtype=float)
        rate = int(self.attack_num / len(target_id_list))
        # padding target score
        self.logger.debug('every target item\'s num %s' % rate)
        for i in range(len(target_id_list)):
            fake_profiles[i * rate : (i + 1) * rate, target_id_list[i]] = 5
        # padding fillers score
        self.logger.debug(self.n_items)
        filler_pool = list(set(range(self.n_items)) - set(target_id_list))
        filler_sampler = lambda x: np.random.choice(x[0], size=x[1], replace=False)
        sampled_cols = np.reshape(
            np.array(
                [
                    filler_sampler([filler_pool, self.filler_num])
                    for _ in range(self.attack_num)
                ]
            ),
            (-1),
        )
        sampled_rows = [
            j for i in range(self.attack_num) for j in [i] * self.filler_num
        ]
        sampled_values = np.random.normal(
            loc=self.global_mean,
            scale=self.global_std,
            size=(self.attack_num * self.filler_num),
        )
        sampled_values = np.round(sampled_values)
        sampled_values[sampled_values > 5] = 5
        sampled_values[sampled_values < 1] = 1
        fake_profiles[sampled_rows, sampled_cols] = sampled_values
        #
        return fake_profiles

    def input_describe(self):
        return {
            "generate_fake": {
                "target_id_list": (Iterable, [VarDim()]),
            }
        }

    def output_describe(self):
        return {
            "generate_fake": {
                "fake_profile": (np.ndarray, (self.attack_num, self.n_items))
            }
        }


class AverageAttack(BaseAttacker):
    def __init__(self, **config):
        super().__init__()
        kvr = config['dataset'].info_describe()['train_kvr']
        self.n_items = config['dataset'].info_describe()['n_items']

        self.global_mean = np.mean(kvr[:, 2])
        self.global_std = np.std(kvr[:, 2])

        self.item_mean_dict = {}
        self.item_std_dict = {}
        for iid in np.unique(kvr[:, 1]):
            self.item_mean_dict[iid] = np.mean(kvr[kvr[:, 1] == iid][:, 2])
            self.item_std_dict[iid] = np.mean(kvr[kvr[:, 1] == iid][:, 2])
        self.attack_num = config['attack_num']
        self.filler_num = config['filler_num']
        self.logger = get_logger(__name__, level=config['logging_level'])

    @classmethod
    def from_config(cls, **kwargs):
        args = list(MODEL['attacker']['average'])
        user_args = "dataset"
        return super().from_config("average", args, user_args, kwargs)

    def train_step(self, users, positive_items, negative_items, **config):
        pass

    def forward(self, users, items):
        pass

    def generate_fake(self, target_id_list, **config):
        fake_profiles = np.zeros(shape=[self.attack_num, self.n_items], dtype=float)
        # padding target score
        rate = int(self.attack_num / len(target_id_list))
        # padding target score
        for i in range(len(target_id_list)):
            fake_profiles[i * rate : (i + 1) * rate, target_id_list[i]] = 5

        filler_pool = list(set(range(self.n_items)) - set(target_id_list))
        filler_sampler = lambda x: np.random.choice(x[0], size=x[1], replace=False)
        sampled_cols = np.reshape(
            np.array(
                [
                    filler_sampler([filler_pool, self.filler_num])
                    for _ in range(self.attack_num)
                ]
            ),
            (-1),
        )
        sampled_rows = [
            j for i in range(self.attack_num) for j in [i] * self.filler_num
        ]
        # sampled_values = np.random.normal(loc=0, scale=1,
        #                                   size=(self.attack_num * self.filler_num))
        sampled_values = [
            np.random.normal(
                loc=self.item_mean_dict.get(iid, self.global_mean),
                scale=self.item_std_dict.get(iid, self.global_std),
            )
            for iid in sampled_cols
        ]
        sampled_values = np.round(sampled_values)
        sampled_values[sampled_values > 5] = 5
        sampled_values[sampled_values < 1] = 1
        fake_profiles[sampled_rows, sampled_cols] = sampled_values
        #
        return fake_profiles

    def input_describe(self):
        return {
            "generate_fake": {
                "target_id_list": (Iterable, [VarDim()]),
            }
        }

    def output_describe(self):
        return {
            "generate_fake": {
                "fake_profile": (np.ndarray, (self.attack_num, self.n_items))
            }
        }


class SegmentAttack(BaseAttacker):
    def __init__(self, **config):
        super().__init__()
        kvr = config['dataset'].info_describe()['train_kvr']
        self.n_items = config['dataset'].info_describe()['n_items']

        self.selected_ids = config['selected_ids']
        self.attack_num = config['attack_num']
        self.filler_num = config['filler_num']
        self.logger = get_logger(__name__, level=config['logging_level'])

    @classmethod
    def from_config(cls, **kwargs):
        args = list(MODEL['attacker']['segment'])
        user_args = "dataset"
        return super().from_config("segment", args, user_args, kwargs)

    def train_step(self, users, positive_items, negative_items, **config):
        pass

    def forward(self, users, items):
        pass

    def generate_fake(self, target_id_list, **config):
        fake_profiles = np.zeros(shape=[self.attack_num, self.n_items], dtype=float)
        # padding target score
        rate = int(self.attack_num / len(target_id_list))
        # padding target score
        for i in range(len(target_id_list)):
            fake_profiles[i * rate : (i + 1) * rate, target_id_list[i]] = 5
        # padding selected score
        fake_profiles[:, self.selected_ids] = 5
        # padding fillers score
        filler_pool = list(
            set(range(self.n_items)) - set(target_id_list) - set(self.selected_ids)
        )
        filler_sampler = lambda x: np.random.choice(x[0], size=x[1], replace=False)
        sampled_cols = np.reshape(
            np.array(
                [
                    filler_sampler([filler_pool, self.filler_num])
                    for _ in range(self.attack_num)
                ]
            ),
            (-1),
        )
        sampled_rows = [
            j for i in range(self.attack_num) for j in [i] * self.filler_num
        ]
        sampled_values = np.ones_like(sampled_rows)
        fake_profiles[sampled_rows, sampled_cols] = sampled_values
        #
        return fake_profiles

    def input_describe(self):
        return {
            "generate_fake": {
                "target_id_list": (Iterable, [VarDim()]),
            }
        }

    def output_describe(self):
        return {
            "generate_fake": {
                "fake_profile": (np.ndarray, (self.attack_num, self.n_items))
            }
        }


class BandwagonAttack(BaseAttacker):
    def __init__(self, **config):
        super().__init__()
        kvr = config['dataset'].info_describe()['train_kvr']

        self.global_mean = np.mean(kvr[:, 2])
        self.global_std = np.std(kvr[:, 2])

        train_data_df = pd.DataFrame(kvr)
        train_data_df.columns = ['user_id', 'item_id', 'rating']
        self.n_items = config['dataset'].info_describe()['n_items']

        self.selected_ids = config['selected_ids']
        self.attack_num = config['attack_num']
        self.filler_num = config['filler_num']
        if len(self.selected_ids) == 0:
            sorted_item_pop_df = (
                train_data_df.groupby('item_id')
                .agg('count')
                .sort_values('user_id')
                .index[::-1]
            )
            self.selected_ids = sorted_item_pop_df[:11].to_list()
        self.logger = get_logger(__name__, level=config['logging_level'])

    @classmethod
    def from_config(cls, **kwargs):
        args = list(MODEL['attacker']['segment'])
        user_args = "dataset"
        return super().from_config("segment", args, user_args, kwargs)

    def train_step(self, users, positive_items, negative_items, **config):
        pass

    def forward(self, users, items):
        pass

    def generate_fake(self, target_id_list, **config):
        fake_profiles = np.zeros(shape=[self.attack_num, self.n_items], dtype=float)
        # padding target score
        # fake_profiles[:, self.target_id] = 5
        rate = int(self.attack_num / len(target_id_list))
        # padding target score
        for i in range(len(target_id_list)):
            fake_profiles[i * rate : (i + 1) * rate, target_id_list[i]] = 5
        # padding selected score
        fake_profiles[:, self.selected_ids] = 5
        # padding fillers score
        filler_pool = list(
            set(range(self.n_items)) - set(target_id_list) - set(self.selected_ids)
        )
        filler_sampler = lambda x: np.random.choice(x[0], size=x[1], replace=False)
        sampled_cols = np.reshape(
            np.array(
                [
                    filler_sampler([filler_pool, self.filler_num])
                    for _ in range(self.attack_num)
                ]
            ),
            (-1),
        )
        sampled_rows = [
            j for i in range(self.attack_num) for j in [i] * self.filler_num
        ]
        sampled_values = np.random.normal(
            loc=self.global_mean,
            scale=self.global_std,
            size=(self.attack_num * self.filler_num),
        )
        sampled_values = np.round(sampled_values)
        sampled_values[sampled_values > 5] = 5
        sampled_values[sampled_values < 1] = 1
        fake_profiles[sampled_rows, sampled_cols] = sampled_values
        #
        return fake_profiles

    def input_describe(self):
        return {
            "generate_fake": {
                "target_id_list": (Iterable, [VarDim()]),
            }
        }

    def output_describe(self):
        return {
            "generate_fake": {
                "fake_profile": (np.ndarray, (self.attack_num, self.n_items))
            }
        }



from functools import partial
from abc import ABC, abstractmethod
from collections.abc import Iterable

from torch.nn import Module
from ..default import MODEL
from ..utils import (
    strip_str,
    get_logger,
    user_side_check,
    parse_args,
    lazy_init,
    type_if_long,
)
from pprint import pprint
from copy import copy

logger = get_logger(__name__)


class BaseModel(Module, ABC):
    def __new__(cls, *args, **kwargs):
        cls = lazy_init(cls, record_attrs=['_init_config', '_model_name'])
        instan = object.__new__(cls)
        return instan

    @classmethod
    @abstractmethod
    def from_config(cls, scope, name, arg_string, user_args, user_config):
        assert name in MODEL[scope], f"{name} is not on the default {scope} models"
        # if not user_side_check(user_args, user_config):
        #     print("will not I")
        # come from default
        args = parse_args(arg_string)
        user_args = parse_args(user_args)
        try:
            default_config = {k: copy(MODEL[scope][name][k]) for k in args}
        except:
            error_message = f"{args} contain non-legal keys for {cls}!, expect in {list(MODEL[scope][name])}"
            raise KeyError(error_message)
        filter_user_config = {}
        for k, v in user_config.items():
            if k in args or k in user_args:
                filter_user_config[k] = v
            else:
                logger.debug(f"Unexpected key [{k}] for {cls}")
        default_config.update(filter_user_config)

        instan = BaseModel.__new__(cls)
        instan._init_config = default_config
        instan._model_name = name
        instan.__init__(**default_config)
        return instan

    # TODO change describe method to classmethod
    def input_describe(self):
        """Input description for method forward and object_forward"""
        raise NotImplementedError

    def output_describe(self):
        """Output description for method forward and object_forward"""
        raise NotImplementedError

    def info_describe(self):
        return {
            "input_describe": self.input_describe(),
            "output_describe": self.output_describe(),
        }

    @abstractmethod
    def train_step(self, **config):
        """Unlike the forward method, this method should return the results relative to the object function"""
        raise NotImplementedError

    @property
    def model_name(self):
        if hasattr(self, "_model_name"):
            return self._model_name
        return type(self).__name__

    def print_help(self, **kwargs) -> str:
        info_des = self.info_describe()
        assert isinstance(
            info_des, dict
        ), "Wrong return type of info_describe, expected Dict"
        info_des['model_name'] = self.model_name
        for k, v in info_des.items():
            if k in ['input_describe', 'output_describe']:
                continue
            v = type_if_long(v)
            info_des[k] = v
        pprint(info_des)

    def reset(self, **kwargs):
        # inject may change some files
        if hasattr(self, "_init_config"):
            config = copy(self._init_config)
            for k, v in kwargs.items():
                if k in config:
                    config[k] = v
                else:
                    raise ValueError(f"reset arg {k} should be in {list(config)}")
            return type(self).from_config(**config)
        raise ValueError("reset method is only for datasets instantiated from_config")



import torch
from .base import BaseData
import numpy as np
import pandas as pd
import os
from os.path import join
from ..utils import get_logger, VarDim, check_dir_or_make
from ..default import DATASET
from scipy.sparse import csr_matrix


class ExplicitData(BaseData):
    def __init__(
        self,
        path_train,
        path_test,
        path_valid,
        header,
        sep,
        threshold,
        logging_level,
        **config,
    ):
        self.path_train = path_train
        self.path_test = path_test
        self.path_valid = path_valid
        self.header = header if header is not None else ['user_id', 'item_id', 'rating']
        self.sep = sep
        self.threshold = threshold
        self.logger = get_logger(f"{__name__}:{self.dataset_name}", level=logging_level)
        self._mode = "train"
        self.config = config
        self.remap_enable = config['remap_enable']
        self._load_data()

    def remap_kvr(self, kvr):
        for i in range(len(kvr)):
            kvr[i, 0] = self.user_map[kvr[i, 0]]
        return kvr

    def _load_data(self):
        self.n_users = 0
        self.n_items = 0

        for attr, default in zip(
            ['train_dict', 'valid_dict', 'test_dict'],
            [self.path_train, self.path_valid, self.path_test],
        ):
            if self.config[attr] is not None:
                self.logger.debug(f"Init {attr} from a passed dict")
                setattr(self, attr, self.config[attr])
                continue
            maybe_cache = join(
                self.config['cache_dir'],
                f"{self.dataset_name}_explicit_{attr}.npy",
            )
            if os.path.exists(maybe_cache) and self.config['if_cache']:
                self.logger.info(f"loading cached {attr}")
                setattr(self, attr, np.load(maybe_cache))
            else:
                setattr(self, attr, self.load_file_as_np(default))
                check_dir_or_make(self.config['cache_dir'])
                if self.config['if_cache']:
                    np.save(maybe_cache, getattr(self, attr))

        if self.remap_enable:
            unique_user = np.unique(
                np.concatenate(
                    [self.train_dict[:, 0], self.test_dict[:, 0], self.valid_dict[:, 0]]
                )
            )
            self.user_map = {unique_user[i]: i for i in range(len(unique_user))}
            self.train_dict = self.remap_kvr(self.train_dict)
            self.test_dict = self.remap_kvr(self.test_dict)
            self.valid_dict = self.remap_kvr(self.valid_dict)

        self.n_users = (
            max(
                self.train_dict[:, 0].max(),
                self.valid_dict[:, 0].max(),
                self.test_dict[:, 0].max(),
            )
            + 1
        )
        self.n_items = (
            max(
                self.train_dict[:, 1].max(),
                self.valid_dict[:, 1].max(),
                self.test_dict[:, 1].max(),
            )
            + 1
        )
        self.n_users = int(self.n_users)
        self.n_items = int(self.n_items)

        self.train_size = len(self.train_dict)
        self.valid_size = len(self.valid_dict)
        self.test_size = len(self.test_dict)

        self.train_mat = self.to_matrix(self.train_dict, self.n_users, self.n_items)

    def load_file_as_np(self, file):
        self.logger.debug("\nload data from %s ..." % file)

        train_data = pd.read_csv(file, engine='python', sep=self.sep)
        train_data = train_data.loc[:, self.header]

        # self.n_users = max(self.n_users, max(train_data.user_id.unique()))
        # self.n_items = max(self.n_items, max(train_data.item_id.unique()))

        train_data = train_data.to_numpy()
        return train_data

    def to_matrix(self, kv_array, n_users, n_items):
        row, col, rating = kv_array[:, 0], kv_array[:, 1], kv_array[:, 2]
        row = row.astype("int64")
        col = col.astype("int64")
        rating = rating.astype("float32")
        # implicit_rating = (rating >= self.threshold).astype("int")

        matrix = csr_matrix((rating, (row, col)), shape=(n_users, n_items)).toarray()
        return matrix

    @classmethod
    def from_config(cls, name, **user_config):
        args = list(DATASET['explicit'][name])
        return super().from_config("explicit", name, args, user_config)

    def batch_describe(self):
        if self.mode() == 'train':
            return {
                'users': (
                    torch.int64,
                    (VarDim(max=self.config['batch_size'], comment="batch size")),
                ),
                'users_mat': (
                    torch.float32,
                    (
                        VarDim(max=self.config['batch_size'], comment="batch size"),
                        self.n_items,
                    ),
                ),
            }

    def info_describe(self):
        infos = {
            "n_users": self.n_users,
            "n_items": self.n_items,
            "train_interactions": self.train_size,
            "valid_interactions": self.valid_size,
            "test_interactions": self.test_size,
            "train_kvr": self.train_dict,
            "train_mat": self.train_mat,
            "batch_describe": self.batch_describe(),
        }
        if self.remap_enable:
            infos['user_map'] = self.user_map
        return infos

    def mode(self) -> str:
        return self._mode

    def generate_batch(self, **config):
        user_filter = config.get("user_filter", None)
        if self.mode() == 'train':
            batch_size = self.config['batch_size']
            if user_filter is not None:
                available_idx = user_filter(train_mat=self.train_mat)
            else:
                available_idx = list(range(len(self.train_mat)))
            available_idx = np.random.permutation(available_idx)
            total_batch = (len(available_idx) + batch_size - 1) // batch_size
            for b in range(total_batch):
                batch_set_idx = available_idx[b * batch_size : (b + 1) * batch_size]
                real_profiles = self.train_mat[batch_set_idx, :].astype('float')

                yield {
                    "users": torch.tensor(batch_set_idx, dtype=torch.int64).to(
                        self.config['device']
                    ),
                    "users_mat": torch.tensor(real_profiles, dtype=torch.float32).to(
                        self.config['device']
                    ),
                }

    def switch_mode(self, mode):
        assert mode in ["train", "test", "validate"]
        self._mode = mode

    def inject_data(self, mode, data):
        # TODO add injection for explicit data
        return super().inject_data(mode, data)

    def partial_sample(self, **kwargs) -> 'BaseData':
        assert "user_ratio" in kwargs, "Expect to have [user_ratio]"
        user_ratio = kwargs['user_ratio']

        users = np.unique(self.train_dict[:, 0])
        np.random.shuffle(users)
        left_users = users[: int(len(users) * user_ratio)]

        left_index = np.zeros(len(self.train_dict), dtype="bool")
        for u in left_users:
            left_index = np.logical_or(left_index, self.train_dict[:, 0] == u)
        new_train_dict = self.train_dict[left_index]

        left_index = np.zeros(len(self.test_dict), dtype="bool")
        for u in left_users:
            left_index = np.logical_or(left_index, self.test_dict[:, 0] == u)
        new_test_dict = self.test_dict[left_index]

        left_index = np.zeros(len(self.valid_dict), dtype="bool")
        for u in left_users:
            left_index = np.logical_or(left_index, self.valid_dict[:, 0] == u)
        new_valid_dict = self.valid_dict[left_index]

        return self.reset(
            train_dict=new_train_dict,
            test_dict=new_test_dict,
            valid_dict=new_valid_dict,
            remap_enable=True,
            if_cache=False,
        )



from ..base import BaseModel


class BaseVictim(BaseModel):
    @classmethod
    def from_config(cls, name, arg_string, user_args, user_config):
        return super().from_config("victim", name, arg_string, user_args, user_config)



"""dataset's returned information should have the common name:
- n_users: the number of the users in dataset
- n_items: the number of the items in dataset
- train_interactions: the interactions number in the dataste
TODO: finish the common fields
.. note::
    _description_
"""
from . import explicit, implicit
from .base import BaseData


factories = {"implicit": implicit.ImplicitData, "explicit": explicit.ExplicitData}


def from_config(scope, *args, **kwargs):
    return factories[scope].from_config(*args, **kwargs)



import os
from .utils import (
    root_path,
    get_logger,
    check_path_or_report,
    dict2list_table,
    check_dir_or_make,
    use_dir,
    download_unzip_to_dir,
)
import torch
from tabulate import tabulate
import logging

_logger = get_logger(__name__)

DATA_ROOT = os.environ.get("RECAD_DIR", root_path())
DATASET_ROOT = os.path.join(DATA_ROOT, "data")
DEVICE = torch.device('cuda' if torch.cuda.is_available() else "cpu")
_logger.info(f"data dir located at {DATA_ROOT}")
SEED = 2023


def _decorate_config(D, k, v, depth):
    if depth == 1:
        D[k] = v
        return
    else:
        for p in D.values():
            _decorate_config(p, k, v, depth - 1)


_Remote_Data_Urls = {
    'yelp': 'https://drive.google.com/uc?export=download&id=1M2bdsOgobgPOtHjH6enBX_m8WtntyHX8',
    'RentTheRunway': 'https://drive.google.com/uc?export=download&id=1jh-W96_DMsFJAaOD9PD8mYmvyDGZ_1vt',
    'ratebeer': 'https://drive.google.com/uc?export=download&id=1G5jEjadW2gbVzJySr9KZrldA1JBpHvLv',
    'ModCloth': 'https://drive.google.com/uc?export=download&id=1OkjHrkWGmXMUERP1aS_e6ZRH7LWVehY9',
    'ml1m': 'https://drive.google.com/uc?export=download&id=1eN1XJvXY27sNQAhTI5qnZlDPc76yIcYn',
    'game': 'https://drive.google.com/uc?export=download&id=1I4EJnBHH73woviP4KC_2-Yn22CqGq3cR',
    'food': 'https://drive.google.com/uc?export=download&id=1KWEQOe5bo6KjeMVFTbJst6m1Ci0iOP5X',
    'epinions': 'https://drive.google.com/uc?export=download&id=1RZ2HAunxinb_fEwd5011D5C_-Bh0FZ2a',
    'dianping': 'https://drive.google.com/uc?export=download&id=1_AlPTV605wnX9Oc_aPmzQlzsx_BGqk6-',
    'book-crossing': 'https://drive.google.com/uc?export=download&id=13UVgWEo0uT6r96WvyyQZRD7y7mDKDVnh',
    'BeerAdvocate': 'https://drive.google.com/uc?export=download&id=1h9jZdyP0FZ-Ql2p00_bAfafT-38r45D_',
    'dev': 'https://github.com/gusye1234/recad/raw/main/data/dev.zip',
}


DATASET = {
    "implicit": {
        name: {
            "path_train": check_path_or_report(DATASET_ROOT, name, f"{name}_train.csv"),
            "path_valid": check_path_or_report(DATASET_ROOT, name, f"{name}_valid.csv"),
            "path_test": check_path_or_report(DATASET_ROOT, name, f"{name}_test.csv"),
            "test_batch_size": 400,
            "A_split": False,
            "A_n_fold": 100,
            "pairwise_batch_size": 1024,
            "pointwise_batch_size": 1024,
            "sample": "pairwise",
            "negative_ratio": 4,
            "need_graph": True,
            "rating_filter": 4,
        }
        for name in list(_Remote_Data_Urls)
    },
    "explicit": {
        name: {
            "path_train": check_path_or_report(DATASET_ROOT, name, f"{name}_train.csv"),
            "path_test": check_path_or_report(DATASET_ROOT, name, f"{name}_test.csv"),
            "path_valid": check_path_or_report(DATASET_ROOT, name, f"{name}_test.csv"),
            "batch_size": 256,
            'header': None,
            'sep': ',',
            'threshold': 4,
            "sample": "row",
        }
        for name in list(_Remote_Data_Urls)
    },
}

_decorate_config(DATASET, "logging_level", logging.INFO, 3)
_decorate_config(DATASET, "train_dict", None, 3)
_decorate_config(DATASET, "valid_dict", None, 3)
_decorate_config(DATASET, "test_dict", None, 3)
_decorate_config(DATASET, "remap_enable", False, 3)  # for partial sample
_decorate_config(DATASET, "device", DEVICE, 3)
_decorate_config(DATASET, "if_cache", False, 3)
_decorate_config(DATASET, "cache_dir", use_dir(".", "generated"), 3)


def dataset_print_help(name):
    for k, v in DATASET.items():
        if name in v:
            print(k)
            print(tabulate(dict2list_table(DATASET[k][name]), disable_numparse=True))


def print_datasets():
    print({k: list(v) for k, v in DATASET.items()})


MODEL = {
    "victim": {
        "lightgcn": {
            "latent_dim_rec": 128,
            "lightGCN_n_layers": 3,
            "A_split": False,
            "pretrain": False,
            "keep_prob": 0.6,
            "dropout": 0.0,
            "lambda": 0.0001,
            "optim": 'adam',
            "lr": 0.001,
        },
        "mf": {
            "factor_num": 3,
            "embedding_size": 128,
            "dropout": 0,
            "optim": 'adam',
            "lr": 0.001,
        },
        "ncf": {
            "factor_num": 32,
            "num_layers": 5,
            "dropout": 0,
            "model": "NeuMF-end",
            "GMF_model": None,
            "MLP_model": None,
            "optim": 'adam',
            "lr": 0.001,
        },
    },
    "attacker": {
        "random": {"attack_num": 50, "filler_num": 36},
        "average": {"attack_num": 50, "filler_num": 36},
        "segment": {
            "attack_num": 50,
            "filler_num": 36,
            "selected_ids": [
                1153,
                2201,
                1572,
                836,
                523,
                849,
                1171,
                344,
                857,
                1213,
                1535,
            ],
        },
        "bandwagon": {
            "attack_num": 50,
            "filler_num": 36,
            "selected_ids": [],
        },
        "aush": {
            "attack_num": 50,
            "filler_num": 36,
            "lr_g": 0.01,
            "lr_d": 0.001,
            "optim_g": 'adam',
            "optim_d": 'adam',
            "selected_ids": [62],
            "ZR_ratio": 0.2,
        },
        "aia": {
            "attack_num": 50,
            "filler_num": 36,
            "lr_g": 0.01,
            "lr_d": 0.001,
            "optim_g": 'adam',
            "optim_d": 'adam',
            "surrogate_model": "WMF",
            "epoch_s": 50,
            "unroll_steps_s": 1,
            "hidden_dim_s": 16,
            "lr_s": 1e-2,
            "weight_decay_s": 1e-5,
            "batch_size_s": 16,
            "weight_pos_s": 1.0,
            "weight_neg_s": 0.0,
            "selected_ids": [62],
        },
        "aushplus": {
            "attack_num": 50,
            "pretrain_epoch_g": 1,
            "pretrain_epoch_d": 5,
            "epoch_gan_d": 5,
            "epoch_gan_g": 1,
            "epoch_surrogate": 50,
            "filler_num": 36,
            "lr_g": 0.01,
            "lr_d": 0.001,
            "optim_g": 'adam',
            "optim_d": 'adam',
            "surrogate_model": "WMF",
            "epoch_s": 50,
            "unroll_steps_s": 1,
            "hidden_dim_s": 16,
            "lr_s": 1e-2,
            "weight_decay_s": 1e-5,
            "batch_size_s": 16,
            "weight_pos_s": 1.0,
            "weight_neg_s": 0.0,
            "selected_ids": [62],
        },
    },
}

_decorate_config(MODEL, "logging_level", logging.INFO, 3)
_decorate_config(MODEL, "device", DEVICE, 3)


def model_print_help(name):
    for k, v in MODEL.items():
        if name in v:
            print(k)
            print(tabulate(dict2list_table(MODEL[k][name]), disable_numparse=True))


def print_models():
    print({k: list(v) for k, v in MODEL.items()})


WORKFLOW = {
    "no defense": {
        "rec_epoch": 400,
        "attack_epoch": 100,
        "target_id_list": [
            0,
        ],
        "filter_num": 4,
        "topks": [10, 20, 50, 100],
    }
}


def workflow_print_help(name):
    for k, v in WORKFLOW.items():
        if name in k:
            print(tabulate(dict2list_table(WORKFLOW[name]), disable_numparse=True))


def print_workflows():
    print(list(WORKFLOW))


_decorate_config(WORKFLOW, "logging_level", logging.INFO, 2)
_decorate_config(WORKFLOW, "device", DEVICE, 2)
_decorate_config(WORKFLOW, "cache_dir", use_dir(".", "workflows_results"), 2)


def set_device_id(cuda_id):
    device = torch.device(f'cuda:{cuda_id}' if torch.cuda.is_available() else "cpu")
    # TODO
    _decorate_config(MODEL, "device", device, 3)
    _decorate_config(WORKFLOW, "device", device, 2)
    _decorate_config(DATASET, "device", device, 3)


# TODO: update to a zh-accessible storage

# Google drive direct download link, generated from https://sites.google.com/site/gdocs2direct/


def download_dataset(scope, name):
    try:
        data_url = _Remote_Data_Urls[name]
    except KeyError:
        raise KeyError(
            f"remote dataset {name} doesn't exist, please consider modify the name in {list(_Remote_Data_Urls)}"
        )
    if not os.path.exists(DATASET_ROOT):
        os.mkdir(DATASET_ROOT)
    download_unzip_to_dir(data_url, DATASET_ROOT, name)



import torch
from .base import BaseVictim
from torch import nn
from ...utils import user_side_check, get_logger, VarDim
from ...default import MODEL
from ...dataset import BaseData
from ...utils import pick_optim


class LightGCN(BaseVictim):
    def __init__(self, **config):
        super(LightGCN, self).__init__()
        self.config = config
        self.dataset: BaseData = config['dataset']
        self.logger = get_logger(__name__, config['logging_level'])
        self.__init_weight()
        self.optimizer = pick_optim(self.config['optim'])(
            self.parameters(), lr=self.config['lr']
        )

    @classmethod
    def from_config(cls, **kwargs):
        args = list(MODEL['victim']['lightgcn'])
        if kwargs.get("pretrain", False):
            user_args = "dataset user_emb, item_emb"
        else:
            user_args = "dataset"
        return super().from_config("lightgcn", args, user_args, kwargs)

    def __init_weight(self):
        self.num_users = self.dataset.info_describe()['n_users']
        self.num_items = self.dataset.info_describe()['n_items']
        self.Graph = self.dataset.info_describe()['graph']
        self.latent_dim = self.config['latent_dim_rec']
        self.n_layers = self.config['lightGCN_n_layers']
        self.keep_prob = self.config['keep_prob']
        self.A_split = self.config['A_split']
        if self.A_split:
            raise ValueError("A_split is not support in LightGCN yet")
        self.embedding_user = torch.nn.Embedding(
            num_embeddings=self.num_users, embedding_dim=self.latent_dim
        )
        self.embedding_item = torch.nn.Embedding(
            num_embeddings=self.num_items, embedding_dim=self.latent_dim
        )
        if self.config['pretrain'] == 0:
            nn.init.normal_(self.embedding_user.weight, std=0.1)
            nn.init.normal_(self.embedding_item.weight, std=0.1)
            self.logger.debug('use NORMAL distribution initilizer')
        else:
            self.embedding_user.weight.data.copy_(
                torch.from_numpy(self.config['user_emb'])
            )
            self.embedding_item.weight.data.copy_(
                torch.from_numpy(self.config['item_emb'])
            )
            self.logger.debug('use pretarined data')
        self.f = nn.Sigmoid()
        self.logger.debug(f"lgn is already to go(dropout:{self.config['dropout']})")
        # print("save_txt")

    def __dropout_x(self, x, keep_prob):
        size = x.size()
        index = x.indices().t()
        values = x.values()
        random_index = torch.rand(len(values)) + keep_prob
        random_index = random_index.int().bool()
        index = index[random_index]
        values = values[random_index] / keep_prob
        g = torch.sparse.FloatTensor(index.t(), values, size)
        return g

    def __dropout(self, keep_prob):
        if self.A_split:
            graph = []
            for g in self.Graph:
                graph.append(self.__dropout_x(g, keep_prob))
        else:
            graph = self.__dropout_x(self.Graph, keep_prob)
        return graph

    def computer(self):
        """
        propagate methods for lightGCN
        """
        users_emb = self.embedding_user.weight
        items_emb = self.embedding_item.weight
        all_emb = torch.cat([users_emb, items_emb])
        #   torch.split(all_emb , [self.num_users, self.num_items])
        embs = [all_emb]
        if self.config['dropout']:
            if self.training:
                g_droped = self.__dropout(self.keep_prob)
            else:
                g_droped = self.Graph
        else:
            g_droped = self.Graph

        for layer in range(self.n_layers):
            if self.A_split:
                temp_emb = []
                for f in range(len(g_droped)):
                    temp_emb.append(torch.sparse.mm(g_droped[f], all_emb))
                side_emb = torch.cat(temp_emb, dim=0)
                all_emb = side_emb
            else:
                all_emb = torch.sparse.mm(g_droped, all_emb)
            embs.append(all_emb)
        embs = torch.stack(embs, dim=1)
        # print(embs.size())
        light_out = torch.mean(embs, dim=1)
        users, items = torch.split(light_out, [self.num_users, self.num_items])
        return users, items

    def getUsersRating(self, users):
        all_users, all_items = self.computer()
        users_emb = all_users[users.long()]
        items_emb = all_items
        rating = self.f(torch.matmul(users_emb, items_emb.t()))
        return rating

    def getEmbedding(self, users, pos_items, neg_items):
        all_users, all_items = self.computer()
        users_emb = all_users[users]
        pos_emb = all_items[pos_items]
        neg_emb = all_items[neg_items]
        users_emb_ego = self.embedding_user(users)
        pos_emb_ego = self.embedding_item(pos_items)
        neg_emb_ego = self.embedding_item(neg_items)
        return users_emb, pos_emb, neg_emb, users_emb_ego, pos_emb_ego, neg_emb_ego

    def train_step(self, **config):
        optim = self.optimizer
        total_loss = 0.0
        self.train()
        pbar = config.get('progress_bar', None)
        for idx, dp in enumerate(self.dataset.generate_batch()):
            users = dp['users']
            pos = dp['positive_items']
            neg = dp['negative_items']
            (
                users_emb,
                pos_emb,
                neg_emb,
                userEmb0,
                posEmb0,
                negEmb0,
            ) = self.getEmbedding(users.long(), pos.long(), neg.long())
            reg_loss = (
                (1 / 2)
                * (
                    userEmb0.norm(2).pow(2)
                    + posEmb0.norm(2).pow(2)
                    + negEmb0.norm(2).pow(2)
                )
                / float(len(users))
            )
            pos_scores = torch.mul(users_emb, pos_emb)
            pos_scores = torch.sum(pos_scores, dim=1)
            neg_scores = torch.mul(users_emb, neg_emb)
            neg_scores = torch.sum(neg_scores, dim=1)

            loss = torch.mean(torch.nn.functional.softplus(neg_scores - pos_scores))

            final_loss = loss + self.config['lambda'] * reg_loss
            optim.zero_grad()
            final_loss.backward()
            optim.step()
            total_loss += final_loss.item()
            if pbar:
                pbar.set_description(f"loss {total_loss / (idx + 1):.5f}")
        return (total_loss / (idx + 1),)

    def forward(self, users, items):
        # compute embedding
        all_users, all_items = self.computer()
        # print('forward')
        # all_users, all_items = self.computer()
        users_emb = all_users[users]
        items_emb = all_items[items]
        inner_pro = torch.mul(users_emb, items_emb)
        gamma = torch.sum(inner_pro, dim=1)
        return gamma

    def input_describe(self):
        return {
            "train_step": {
                "users": (torch.int64, (VarDim(comment="batch"))),
                "positive_items": (torch.int64, (VarDim(comment="batch"))),
                "negative_items": (torch.int64, (VarDim(comment="batch"))),
            },
            "forward": {
                "users": (torch.int64, (VarDim(comment="batch"))),
                "items": (torch.int64, (VarDim(comment="batch"))),
            },
        }

    def output_describe(self):
        return {
            "train_step": {
                "loss": (float, []),
            },
            "forward": {
                "unnormalized_scores": (torch.float32, [VarDim(comment="batch")])
            },
        }



from .base import BaseVictim
import torch
from torch import nn
from ...utils import pick_optim, VarDim
from ...default import MODEL


class NCF(BaseVictim):
    def __init__(
        self, factor_num, num_layers, dropout, model, GMF_model, MLP_model, **config
    ):
        super(NCF, self).__init__()
        """
        user_num: number of users;
        item_num: number of items;
        factor_num: number of predictive factors;
        num_layers: the number of layers in MLP model;
        dropout: dropout rate between fully connected layers;
        model: 'MLP', 'GMF', 'NeuMF-end', and 'NeuMF-pre';
        GMF_model: pre-trained GMF weights;
        MLP_model: pre-trained MLP weights.
        """
        self.dataset = config['dataset']
        user_num = self.dataset.info_describe()['n_users']
        item_num = self.dataset.info_describe()['n_items']

        self.dropout = dropout
        self.model = model
        self.GMF_model = GMF_model
        self.MLP_model = MLP_model

        self.embed_user_GMF = nn.Embedding(user_num, factor_num)
        self.embed_item_GMF = nn.Embedding(item_num, factor_num)
        self.embed_user_MLP = nn.Embedding(
            user_num, factor_num * (2 ** (num_layers - 1))
        )
        self.embed_item_MLP = nn.Embedding(
            item_num, factor_num * (2 ** (num_layers - 1))
        )

        MLP_modules = []
        for i in range(num_layers):
            input_size = factor_num * (2 ** (num_layers - i))
            MLP_modules.append(nn.Dropout(p=self.dropout))
            MLP_modules.append(nn.Linear(input_size, input_size // 2))
            MLP_modules.append(nn.ReLU())
        self.MLP_layers = nn.Sequential(*MLP_modules)

        if self.model in ['MLP', 'GMF']:
            predict_size = factor_num
        else:
            predict_size = factor_num * 2
        self.predict_layer = nn.Linear(predict_size, 1)

        self._init_weight_()

        self.optimizer = pick_optim(config['optim'])(self.parameters(), lr=config['lr'])
        self.loss_func = nn.BCEWithLogitsLoss()

    def _init_weight_(self):
        """We leave the weights initialization here."""
        if not self.model == 'NeuMF-pre':
            nn.init.normal_(self.embed_user_GMF.weight, std=0.01)
            nn.init.normal_(self.embed_user_MLP.weight, std=0.01)
            nn.init.normal_(self.embed_item_GMF.weight, std=0.01)
            nn.init.normal_(self.embed_item_MLP.weight, std=0.01)

            for m in self.MLP_layers:
                if isinstance(m, nn.Linear):
                    nn.init.xavier_uniform_(m.weight)
            nn.init.kaiming_uniform_(
                self.predict_layer.weight, a=1, nonlinearity='sigmoid'
            )

            for m in self.modules():
                if isinstance(m, nn.Linear) and m.bias is not None:
                    m.bias.data.zero_()
        else:
            # embedding layers
            self.embed_user_GMF.weight.data.copy_(self.GMF_model.embed_user_GMF.weight)
            self.embed_item_GMF.weight.data.copy_(self.GMF_model.embed_item_GMF.weight)
            self.embed_user_MLP.weight.data.copy_(self.MLP_model.embed_user_MLP.weight)
            self.embed_item_MLP.weight.data.copy_(self.MLP_model.embed_item_MLP.weight)

            # mlp layers
            for (m1, m2) in zip(self.MLP_layers, self.MLP_model.MLP_layers):
                if isinstance(m1, nn.Linear) and isinstance(m2, nn.Linear):
                    m1.weight.data.copy_(m2.weight)
                    m1.bias.data.copy_(m2.bias)

            # predict layers
            predict_weight = torch.cat(
                [
                    self.GMF_model.predict_layer.weight,
                    self.MLP_model.predict_layer.weight,
                ],
                dim=1,
            )
            precit_bias = (
                self.GMF_model.predict_layer.bias + self.MLP_model.predict_layer.bias
            )

            self.predict_layer.weight.data.copy_(0.5 * predict_weight)
            self.predict_layer.bias.data.copy_(0.5 * precit_bias)

    @classmethod
    def from_config(cls, **kwargs):
        args = list(MODEL['victim']['ncf'])
        user_args = "dataset"
        return super().from_config("ncf", args, user_args, kwargs)

    def forward(self, user, item):
        if not self.model == 'MLP':
            embed_user_GMF = self.embed_user_GMF(user)
            embed_item_GMF = self.embed_item_GMF(item)
            output_GMF = embed_user_GMF * embed_item_GMF
        if not self.model == 'GMF':
            embed_user_MLP = self.embed_user_MLP(user)
            embed_item_MLP = self.embed_item_MLP(item)
            interaction = torch.cat((embed_user_MLP, embed_item_MLP), -1)
            output_MLP = self.MLP_layers(interaction)

        if self.model == 'GMF':
            concat = output_GMF
        elif self.model == 'MLP':
            concat = output_MLP
        else:
            concat = torch.cat((output_GMF, output_MLP), -1)

        prediction = self.predict_layer(concat)
        return prediction.view(-1)

    def train_step(self, **config):
        optim = self.optimizer
        total_loss = 0.0
        self.train()
        pbar = config.get('progress_bar', None)
        for idx, dp in enumerate(self.dataset.generate_batch()):
            user = dp['users']
            item = dp['items']
            label = dp['labels']

            pred = self(user, item)
            loss = self.loss_func(pred, label.float())

            optim.zero_grad()
            loss.backward()
            optim.step()

            total_loss += loss.item()
            if pbar:
                pbar.set_description(f"loss: {total_loss/(idx+1):.5f}")
        return (total_loss / (idx + 1),)

    def input_describe(self):
        return {
            "train_step": {
                "users": (torch.int64, (VarDim(comment="batch"))),
                "items": (torch.int64, (VarDim(comment="batch"))),
                "labels": (torch.int64, (VarDim(comment="batch"))),
            },
            "forward": {
                "users": (torch.int64, (VarDim(comment="batch"))),
                "items": (torch.int64, (VarDim(comment="batch"))),
            },
        }

    def output_describe(self):
        return {
            "train_step": {
                "loss": (float, []),
            },
            "forward": {
                "unnormalized_scores": (torch.float32, [VarDim(comment="batch")])
            },
        }



from . import victim, attacker

factories = {
    "victim": {"lightgcn": victim.LightGCN, 'mf': victim.MF, 'ncf': victim.NCF},
    "attacker": {
        "random": attacker.RandomAttack,
        "average": attacker.AverageAttack,
        "segment": attacker.SegmentAttack,
        "bandwagon": attacker.BandwagonAttack,
        "aush": attacker.Aush,
        "aia": attacker.AIA,
        "aushplus": attacker.AushPlus,
    },
}


def from_config(scope, name, **kwargs):
    return factories[scope][name].from_config(**kwargs)



from functools import partial
from typing import Dict, Generator
from abc import ABC, abstractmethod
from collections.abc import Iterable
from ..default import DATASET, download_dataset
from ..utils import strip_str, type_if_long, parse_args, get_logger
from pprint import pprint
from copy import copy
from os.path import exists

logger = get_logger(__name__)


class BaseData(ABC):
    @classmethod
    @abstractmethod
    def from_config(cls, scope, name, arg_string, user_config):
        assert name in DATASET[scope], f"{name} is not on the default datasets"
        args = parse_args(arg_string)
        try:
            default_config = {k: copy(DATASET[scope][name][k]) for k in args}
        except:
            error_message = f"{args} contain non-legal keys for {cls}!, expect in {list(DATASET[scope][name])}"
            raise KeyError(error_message)
        for k in user_config.keys():
            if k not in args:
                logger.debug(f"Unexpected key [{k}] for {cls}, expected in {args}")
        default_config.update(user_config)
        if any(
            [
                not exists(default_config[path_str])
                for path_str in ['path_train', 'path_valid', 'path_test']
            ]
        ):
            if user_config.get("download", True):
                logger.info(
                    f"Some paths of the dataset were missing, triggering download"
                )
                download_dataset(scope, name)
            else:
                logger.info(
                    f"Some paths of the dataset were missing, ignoring it since download=False"
                )

        instan = object.__new__(cls)
        instan._dataset_name = name
        instan._init_config = default_config
        instan.__init__(**default_config)
        return instan

    @abstractmethod
    def info_describe(self) -> Dict:
        """return the information description of this dataset
        for example:
            {
                "n_users": 1000,
                "n_items": 500,
            }
        :raise: NotImplementedError

        :return: a dict of each data field
        """
        raise NotImplementedError

    @abstractmethod
    def mode(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def generate_batch(self, **config) -> Generator:
        raise NotImplementedError

    @abstractmethod
    def switch_mode(self, mode):
        """switch between train, test, validation

        :param mode: in [train, test, valid]
        """
        raise NotImplementedError

    @abstractmethod
    def inject_data(self, data_model, data) -> 'BaseData':
        """inject the attack data

        :param data_model: explicit or implicit, default explicit mode
        :param data: array(fake_users X n_items), each element is a rating

        .. note::
            please notice, the return datasets shouldn't related to the original one
        """
        raise NotImplementedError

    @abstractmethod
    def partial_sample(self, **kwargs) -> 'BaseData':
        """Sample a part of the dataset, to form a new dataset

        .. note::
            please notice, the return datasets shouldn't related to the original one
        """
        raise NotImplementedError

    @property
    def dataset_name(self):
        if hasattr(self, "_dataset_name"):
            return self._dataset_name
        return type(self).__name__

    def reset(self, **kwargs):
        # inject may change some files
        if hasattr(self, "_init_config"):
            config = copy(self._init_config)
            for k, v in kwargs.items():
                if k in config:
                    config[k] = v
                else:
                    raise ValueError(f"reset arg {k} should be in {list(config)}")
            return type(self).from_config(self.dataset_name, **config)
        raise ValueError("reset method is only for datasets instantiated from_config")

    def print_help(self, **kwargs) -> str:
        info_des = self.info_describe()
        assert isinstance(
            info_des, dict
        ), "Wrong return type of info_describe, expected Dict"
        info_des['dataset_name'] = self.dataset_name
        for k, v in info_des.items():
            if k == 'batch_describe':
                continue
            v = type_if_long(v)
            info_des[k] = v
        pprint(info_des)



from os import path
import torch
import pandas as pd
import numpy as np
from .base import BaseWorkflow
from ..dataset import BaseData
from ..model.attacker import BaseAttacker
from ..model.victim import BaseVictim
from ..default import WORKFLOW
from ..utils import use_dir, pick_optim, get_logger, dict2list_table, fmt_tab, tqdm
from collections import OrderedDict
import logging


class Normal(BaseWorkflow):
    def __init__(self, **config):
        self.c = config
        self.saving_dir = use_dir(
            config['cache_dir'],
            "_".join(
                [
                    "no_defense",
                    config['victim_data'].dataset_name,
                    config['attack_data'].dataset_name,
                    config['victim'].model_name,
                    config['attacker'].model_name,
                ]
            ),
        )
        self.attacker: BaseAttacker = config['attacker'].I(
            dataset=config['attack_data']
        )
        self.victim: BaseVictim = config['victim'].I(dataset=config['victim_data'])
        self.victim_data: BaseData = config['victim_data']
        self.logger = get_logger(__name__, level=self.c['logging_level'])

    @classmethod
    def from_config(cls, **kwargs):
        args = list(WORKFLOW['no defense'])
        user_args = "victim_data, attack_data, victim, attacker"
        return super().from_config("no defense", args, user_args, kwargs)

    def input_describe(self):
        return {
            "victim": BaseVictim,
            "victim_data": BaseData,
            "attacker": BaseAttacker,
            "attack_data": BaseData,
        }

    def info_describe(self):
        return {
            'target_id_list': self.c['target_id_list'],
            'input_describe': self.input_describe(),
        }

    def user_item_model_generate(self, model, data_dict, topks, target_ids, device):
        preds = []
        users = []
        items = []
        with torch.no_grad():
            for u, iids in tqdm(data_dict.items(), desc="Inferring"):
                if not len(iids):
                    continue
                tensor_iids = torch.tensor(iids, dtype=torch.int64).to(device)
                us = torch.full_like(tensor_iids, u)
                pred = model(us, tensor_iids).cpu().numpy().tolist()

                preds.extend(pred)
                users.extend([u] * len(iids))
                items.extend(iids)
        pred_results = pd.DataFrame(
            {'user_id': users, 'item_id': items, 'rating': preds}
        )

        topks_array = np.zeros(
            [len(pred_results.user_id.unique()) * len(target_ids), len(topks) + 2]
        )
        pbar = tqdm(enumerate(pred_results.groupby('user_id')))
        for idx, (user_id, pred_result) in pbar:
            pbar.set_description(f"user {user_id}")
            for bias, target_id in enumerate(target_ids):
                pred_value = pred_result[
                    pred_result.item_id == target_id
                ].rating.values[0]
                sorted_recommend_list = pred_result.sort_values(
                    'rating', ascending=False
                ).item_id.values
                new_line = [user_id, pred_value] + [
                    1 if target_id in sorted_recommend_list[:k] else 0 for k in topks
                ]
                topks_array[idx + bias] = new_line
        return topks_array

    def normal_train(self, **config):
        progress = tqdm(range(config['epoch']))
        for ep in progress:
            loss = config['model'].train_step(
                **self.info_describe(), progress_bar=progress
            )
            out_des = config['model'].output_describe()['train_step']

            assert len(loss) == len(
                out_des
            ), f"The output describe is not aligned with the actual output of train_step for {config['model'].model_name}"
            loss_flag = ""
            for i, out in enumerate(out_des.keys()):
                loss_flag = loss_flag + f"{out}: {loss[i]:.4f}|"
            progress.set_description(f"{loss_flag}")

    def normal_evaluate(
        self,
        model: BaseVictim,
        model_fake: BaseVictim,
        dataset: BaseData,
        target_id_list,
        topks,
    ):
        train_dict = dataset.info_describe()['train_dict']
        n_users = dataset.info_describe()['n_users']
        n_items = dataset.info_describe()['n_items']
        for m in [model, model_fake]:
            assert (
                len(m.input_describe()['forward']) == 2
            ), "Expect forward only need two inputs"
            assert (
                "users" in m.input_describe()['forward']
            ), "Expect to have the users input in forward method"
            assert (
                "items" in m.input_describe()['forward']
            ), "Expect to have the items input in forward method"

        filtered_dict = {k: set(v) for k, v in train_dict.items()}
        full_items = set(range(n_items))
        for k in train_dict:
            existed = False
            for target in target_id_list:
                if target in filtered_dict[k]:
                    del filtered_dict[k]
                    existed = True
                    break
            if not existed:
                filtered_dict[k] = list(full_items - filtered_dict[k])
        self.logger.debug("Done filtered")
        pred_results = self.user_item_model_generate(
            model, filtered_dict, topks, target_id_list, self.c['device']
        )
        pred_results_fake = self.user_item_model_generate(
            model_fake, filtered_dict, topks, target_id_list, self.c['device']
        )
        assert np.allclose(
            pred_results[:, 0], pred_results_fake[:, 0]
        ), "Users are not aligned"

        results = OrderedDict()
        results['pred_shift'] = np.mean(pred_results_fake[:, 1] - pred_results[:, 1])
        for i, k in enumerate(topks):
            results[f"HR@{k}"] = np.mean(pred_results[:, 2 + i])
            results[f"HR@{k} after attack"] = np.mean(pred_results_fake[:, 2 + i])
        print(fmt_tab(dict2list_table(results)))

    def execute(self):
        self.logger.info(
            f"Normal attacking, with dataset {self.victim_data.dataset_name}, victim model {self.victim.model_name}, attack model {self.attacker.model_name}, on device {self.c['device']}"
        )

        self.victim = self.victim.to(self.c['device'])
        self.attacker = self.attacker.to(self.c['device'])

        self.logger.info("Step 1. training a recommender")
        self.normal_train(
            **{
                'model': self.victim,
                'epoch': self.c['rec_epoch'],
                'dataset': self.victim_data,
            }
        )

        self.logger.info("Step 2. training a attacker")
        if "train_step" in self.attacker.input_describe():
            self.normal_train(
                **{
                    'model': self.attacker,
                    'epoch': self.c['attack_epoch'],
                    'dataset': self.c['attack_data'],
                }
            )
        else:
            self.logger.info(
                f"Skip attacker training, since {self.attacker.model_name} didn't require it"
            )

        fake_array = self.attacker.generate_fake(**self.info_describe())
        self.logger.info(
            f"Step 3. injecting fake data({tuple(fake_array.shape)}) and re-train the recommender"
        )
        self.logger.debug(f"{fake_array.shape}")
        fake_dataset = self.victim_data.inject_data(
            "explicit", fake_array, filter_num=self.c['filter_num']
        )
        self.logger.debug("After injection of the fake data")
        if self.c['logging_level'] == logging.DEBUG:
            fake_dataset.print_help()

        fake_victim = self.victim.reset().I(dataset=fake_dataset)
        fake_victim = fake_victim.to(self.c['device'])
        self.normal_train(
            **{
                'model': fake_victim,
                'epoch': self.c['rec_epoch'],
                'dataset': fake_dataset,
            }
        )

        # TODO add evaluate
        self.logger.debug(
            f"targets: {self.c['target_id_list']}, topks: {self.c['topks']}"
        )
        self.normal_evaluate(
            self.victim,
            fake_victim,
            self.victim_data,
            self.c['target_id_list'],
            topks=self.c['topks'],
        )



import torch
from torch import nn
import torch.nn.functional as F
import numpy as np
from .base import BaseAttacker
from .aia import AIA, RecsysGenerator, partial_update, aia_attack_loss
from ...default import MODEL
from ...utils import pick_optim, VarDim


class AushPlus(AIA):
    def __init__(self, **config) -> None:
        super().__init__(**config)
        self.pretrained = False

    @classmethod
    def from_config(cls, **kwargs):
        print(AushPlus.__bases__)
        args = list(MODEL['attacker']['aushplus'])
        user_args = "dataset"
        return super(AIA, cls).from_config("aushplus", args, user_args, kwargs)

    def build_network(self):
        sampled_idx = np.random.choice(range(self.n_users), self.attack_num)
        templates = self.train_array[sampled_idx]
        for idx, template in enumerate(templates):
            fillers = np.where(template)[0]
            np.random.shuffle(fillers)
            for iid in fillers[self.filler_num :]:
                templates[idx][iid] = 0.0
        self.real_template = torch.tensor(templates).type(torch.float).to(self.device)
        self.netG = DiscretGenerator_AE_1(self.device, p_dims=[self.n_items, 125]).to(
            self.device
        )
        self.G_optimizer = pick_optim(self.config['optim_g'])(
            self.netG.parameters(), lr=self.config['lr_g']
        )

        self.netD = Discriminator(input_dim=self.n_items).to(self.device)
        self.D_optimizer = pick_optim(self.config['optim_g'])(
            self.netD.parameters(), lr=self.config['lr_d']
        )

    def input_describe(self):
        return {"train_step": {"target_id_list": (list, VarDim())}}

    def output_describe(self):
        return {"train_step": {"g_adv": (float, []), "g_rec": (float, [])}}

    def train_G(self, adv=True, attack=True, target_id_list=[]):
        self.netG.train()

        fake_tensor_distribution, fake_tensor = self.netG(self.real_template)

        G_adv_loss = torch.tensor(0.0)
        if adv:
            G_adv_loss = nn.BCELoss(reduction='mean')(
                self.netD(fake_tensor),
                torch.ones(fake_tensor.shape[0], 1).to(self.device),
            )

        real_labels_flatten = self.real_template.flatten().type(torch.long)

        MSELoss = nn.MSELoss()(
            fake_tensor.flatten()[real_labels_flatten > 0],
            real_labels_flatten[real_labels_flatten > 0],
        )

        G_attack_loss = torch.tensor(0.0).to(self.device)
        if attack:
            sur_predictions = self.get_sur_predictions(fake_tensor)
            for target_id in target_id_list:
                target_users = np.where(self.train_array[:, target_id] == 0)[0]
                # target_users 所有对target id 没有评分的 users

                attack_target = np.zeros((len(target_users), self.n_items))
                # attack_target 张成的一个 矩阵 (0)
                attack_target[:, target_id] = 1.0
                # 另矩阵中全部的 target_id == 1.0
                attack_target = (
                    torch.from_numpy(attack_target).type(torch.float32).to(self.device)
                )
                higher_mask = (
                    sur_predictions[target_users]
                    >= (sur_predictions[target_users, target_id].reshape([-1, 1]))
                ).float()

                G_loss_sub = aia_attack_loss(
                    logits=sur_predictions[target_users] * higher_mask,
                    labels=attack_target,
                ).mean()
                G_attack_loss += G_loss_sub
            G_attack_loss = G_attack_loss / 10

        G_loss = G_adv_loss + G_attack_loss

        partial_update(G_loss, self.G_optimizer)
        self.netG.eval()

        mean_score = fake_tensor[fake_tensor > 0].mean().item()
        return (G_loss.item(), MSELoss.item(), G_attack_loss.item(), mean_score)

    def train_D(self):
        self.netD.train()

        _, fake_tensor = self.netG(self.real_template)
        fake_tensor = fake_tensor.detach()

        D_loss_list = []
        loss_bce = nn.BCELoss()
        for dp in self.dataset.generate_batch():
            real_tensor = dp['users_mat']
            real_tensor = real_tensor.to(self.device)[: self.attack_num]
            # forward
            self.D_optimizer.zero_grad()
            D_real = self.netD(real_tensor)
            D_fake = self.netD(fake_tensor)
            # loss
            D_real_loss = loss_bce(D_real, torch.ones_like(D_real).to(self.device))
            D_fake_loss = loss_bce(D_fake, torch.zeros_like(D_fake).to(self.device))
            D_loss = D_real_loss + D_fake_loss
            # backward
            D_loss.backward()
            self.D_optimizer.step()

            D_loss_list.append(D_loss.item())
            #
            # break
        self.netD.eval()

        return np.mean(D_loss_list)

    def pretrain_G(self):
        self.netG.train()
        G_loss_list = []
        loss_cross = nn.CrossEntropyLoss()
        loss_mse = nn.MSELoss()
        for dp in self.dataset.generate_batch():
            real_tensor = dp['users_mat']
            # forward
            fake_tensor_distribution, fake_tensor = self.netG(real_tensor)
            # crossEntropy loss
            real_labels_flatten = real_tensor.flatten().type(torch.long)
            fake_logits_flatten = fake_tensor_distribution.reshape([-1, 5])
            G_rec_loss = loss_cross(
                fake_logits_flatten[real_labels_flatten > 0],
                real_labels_flatten[real_labels_flatten > 0] - 1,
            )
            G_loss = G_rec_loss
            MSELoss = loss_mse(
                fake_tensor.flatten()[real_labels_flatten > 0],
                real_labels_flatten[real_labels_flatten > 0],
            )
            # backword
            partial_update(G_loss, self.G_optimizer)
            G_loss_list.append(G_loss.item())
        self.netG.eval()
        return np.mean(G_loss_list)

    def train_step(self, **config):
        target_id_list = config['target_id_list']
        if not self.pretrained:
            for i in range(self.config['pretrain_epoch_g']):
                self.pretrain_G()
            for i in range(self.config['pretrain_epoch_g']):
                self.train_D()
            self.pretrained = True
        for _ in range(self.config["epoch_gan_d"]):
            self.train_D()
        for _ in range(self.config["epoch_gan_g"]):
            _, _, G_adv_loss, _ = self.train_G(
                adv=True, attack=False, target_id_list=target_id_list
            )
            _, fake_tensor = self.netG(self.real_template)
        for epoch_surrogate in range(self.config['epoch_surrogate']):
            _, _, G_rec_loss, _ = self.train_G(
                adv=False, attack=True, target_id_list=target_id_list
            )

        return (G_adv_loss, G_rec_loss)

    def generate_fake(self, **config):
        target_id_list = config['target_id_list']
        with torch.no_grad():
            _, fake_tensor = self.netG(self.real_template)
            rate = int(self.attack_num / len(target_id_list))
            for i in range(len(target_id_list)):
                fake_tensor[i * rate : (i + 1) * rate, target_id_list[i]] = 5
        return fake_tensor.detach().cpu().numpy()


class HeaviTanh(torch.autograd.Function):
    """
    Approximation of the heaviside step function as
    h(x,k) = \frac{1}{2} + \frac{1}{2} \text{tanh}(k x)
    """

    @staticmethod
    def forward(ctx, x, k):
        ctx.save_for_backward(x, k)

        def heaviside(data):
            """
            A `heaviside step function <https://en.wikipedia.org/wiki/Heaviside_step_function>`_
            that truncates numbers <= 0 to 0 and everything else to 1.
            .. math::
                H[n]=\\begin{cases} 0, & n <= 0, \\ 1, & n \g 0, \end{cases}
            """
            return torch.where(
                data <= torch.zeros_like(data),
                torch.zeros_like(data),
                torch.ones_like(data),
            )

        return heaviside(x)  # 0.5 + 0.5 * torch.tanh(k * x)

    @staticmethod
    def backward(ctx, dy):
        (
            x,
            k,
        ) = ctx.saved_tensors
        dtanh = 1 - (x * k).tanh().pow(2)
        return dy * dtanh, None


class BaseGenerator(nn.Module):
    def __init__(self, device, input_dim):
        super(BaseGenerator, self).__init__()
        #
        self.input_dim = input_dim
        self.device = device

        """helper_tensor"""

        self.epsilon = torch.tensor(1e-4).to(self.device)  # 计算boundary
        self.helper_tensor = torch.tensor(2.5).to(device)
        pass

    def project(self, fake_tensor):
        fake_tensor.data = torch.round(fake_tensor)
        # fake_tensor.data = torch.where(fake_tensor < 1, torch.ones_like(fake_tensor).to(self.device), fake_tensor)
        fake_tensor.data = torch.where(
            fake_tensor < 0, torch.zeros_like(fake_tensor).to(self.device), fake_tensor
        )
        fake_tensor.data = torch.where(
            fake_tensor > 5, torch.tensor(5.0).to(self.device), fake_tensor
        )
        #
        return fake_tensor

    def forward(self, input):
        raise NotImplementedError


class BaseDiscretGenerator_1(BaseGenerator):
    def __init__(self, device, input_dim):
        super(BaseDiscretGenerator_1, self).__init__(device, input_dim)

        # self.min_boundary_value = torch.nn.Parameter(torch.rand([self.input_dim]), requires_grad=True)
        self.min_boundary_value = torch.nn.Parameter(
            torch.ones([self.input_dim]), requires_grad=True
        )
        self.register_parameter("min_boundary_value", self.min_boundary_value)

        # self.interval_lengths = torch.nn.Parameter(torch.rand([self.input_dim, 3]), requires_grad=True)
        self.interval_lengths = torch.nn.Parameter(
            torch.ones([self.input_dim, 3]), requires_grad=True
        )
        self.register_parameter("interval_lengths", self.interval_lengths)

        pass

    def forward(self, input):
        # fake_tensor = (self.main(input) * self.helper_tensor) + self.helper_tensor
        # # project
        # fake_dsct_distribution, fake_dsct_value = self.project(fake_tensor)
        # return fake_dsct_value
        raise NotImplementedError

    def project_old(self, fake_tensor):
        boundary_values = self.get_boundary_values()

        fake_dsct_distribution = []

        for iid in range(self.input_dim):
            cur_dsct_distribution = []
            for rating_dsct in range(5):
                rating_prob = torch.ones(fake_tensor.shape[0]).to(self.device)
                for boundary_idx in range(4):
                    rating_prob *= self.is_in_interval(
                        rating_dsct,
                        boundary_idx,
                        fake_tensor[:, iid],
                        boundary_values[iid][boundary_idx],
                    )
                cur_dsct_distribution += [rating_prob]
            fake_dsct_distribution += [
                torch.cat([torch.unsqueeze(p, 1) for p in cur_dsct_distribution], 1)
            ]
        fake_dsct_distribution = torch.cat(
            [torch.unsqueeze(p, 1) for p in fake_dsct_distribution], 1
        )

        fake_dsct_value = torch.matmul(
            fake_dsct_distribution,
            torch.tensor(np.arange(1.0, 6.0)).type(torch.float32).to(self.device),
        )
        return fake_dsct_distribution, fake_dsct_value

    def project(self, fake_tensor):
        Heaviside = HeaviTanh.apply

        boundary_values = self.get_boundary_values()

        cnt_ratings = fake_tensor.flatten()
        iids = (
            np.expand_dims(np.arange(self.input_dim), 0)
            .repeat(fake_tensor.shape[0], axis=0)
            .flatten()
        )
        boundary_values_per_rating = boundary_values[iids]

        def _project_helper(ratings, boundary_values_input):
            def get_target_dst_rating_prob(
                target_dst_rating, input_cnt_rating, boundary_values, device
            ):
                # boundary_values = boundary_values.reshape([-1, 4])
                # input_cnt_rating = input_cnt_rating.reshape([-1])
                rating_prob = torch.ones(input_cnt_rating.shape[0]).to(self.device)
                for boundary_idx in range(4):
                    """
                    :param target_dst_rating: r_i_j
                    :param boundary_idx: k
                    :param input_cnt_rating: a_i_j
                    :param boundary_value: b_j_k
                    :return:
                    """
                    #
                    p_1 = torch.sign(
                        target_dst_rating - boundary_idx - torch.tensor(0.5).to(device)
                    )
                    #
                    p_2 = input_cnt_rating - boundary_values[:, boundary_idx]
                    #
                    rating_prob *= Heaviside(p_1 * p_2, torch.tensor(1.0).to(device))
                return rating_prob

            cur_dsct_distribution = []
            for rating_dsct in range(5):
                p = get_target_dst_rating_prob(
                    rating_dsct, ratings, boundary_values_input, self.device
                )
                cur_dsct_distribution += [p]
            dsct_distribution = torch.cat(
                [torch.unsqueeze(p, 1) for p in cur_dsct_distribution], 1
            )
            return dsct_distribution

        fake_dsct_distribution = _project_helper(
            cnt_ratings, boundary_values_per_rating
        ).reshape([-1, self.input_dim, 5])

        fake_dsct_value = torch.matmul(
            fake_dsct_distribution,
            torch.tensor(np.arange(1.0, 6.0)).type(torch.float32).to(self.device),
        )
        return fake_dsct_distribution, fake_dsct_value

    def get_boundary_values(self):
        boundary_values = torch.zeros([self.input_dim, 4]).to(self.device)
        boundary_values[:, 0] = self.min_boundary_value
        for i in range(1, 4):
            cur_interval_length = (
                torch.relu(self.interval_lengths[:, i - 1]) + self.epsilon
            )
            boundary_values[:, i] = boundary_values[:, i - 1] + cur_interval_length
        return boundary_values

    def is_in_interval(self, rating_dsct, boundary_idx, rating_cnt, boundary_value):
        tensor_aux_0_5 = torch.tensor(0.5).to(self.device)
        tensor_aux_1 = torch.tensor(1.0).to(self.device)
        Heaviside = HeaviTanh.apply
        """
    
        :param rating_dsct: r_i_j
        :param boundary_idx: k
        :param rating_cnt: a_i_j
        :param boundary_value: b_j_k
        :return:
        """
        #
        p_1 = torch.sign(rating_dsct - boundary_idx - tensor_aux_0_5)
        #
        p_2 = rating_cnt - boundary_value
        #
        return Heaviside(p_1 * p_2, tensor_aux_1)


class DiscretGenerator_AE_1(BaseDiscretGenerator_1):
    def __init__(self, device, p_dims, q_dims=None):
        super(DiscretGenerator_AE_1, self).__init__(device, input_dim=p_dims[0])
        self.p_dims = p_dims
        if q_dims:
            assert (
                q_dims[0] == p_dims[-1]
            ), "In and Out dimensions must equal to each other"
            assert (
                q_dims[-1] == p_dims[0]
            ), "Latent dimension for p- and q- network mismatches."
            self.q_dims = q_dims
        else:
            self.q_dims = p_dims[::-1]

        self.dims = self.p_dims + self.q_dims[1:]
        self.layers = nn.ModuleList(
            [
                nn.Linear(d_in, d_out)
                for d_in, d_out in zip(self.dims[:-1], self.dims[1:])
            ]
        )
        # self.drop = nn.Dropout(dropout)
        self.init_weights()

    def forward(self, input):
        h = F.normalize(input)
        # h = self.drop(h)

        for i, layer in enumerate(self.layers):
            h = layer(h)
            if i != len(self.layers) - 1:
                h = F.relu(h)
            else:
                h = torch.nn.Tanh()(h)

        fake_tensor = (h * self.helper_tensor) + self.helper_tensor
        # project

        fake_dsct_distribution, fake_dsct_value = self.project(fake_tensor)

        sampled_filler = input > 0

        # sampled_filler = (torch.rand(fake_dsct_value.shape) < (90 / 1924)).float()

        # filler_num = np.sum(sampled_filler.detach().cpu().numpy()*(fake_dsct_value.detach().cpu().numpy()>0),1).mean()
        # if filler_num<90:
        return fake_dsct_distribution, fake_dsct_value * sampled_filler

    def init_weights(self):
        for layer in self.layers:
            # Xavier Initialization for weights
            size = layer.weight.size()
            fan_out = size[0]
            fan_in = size[1]
            std = np.sqrt(2.0 / (fan_in + fan_out))
            layer.weight.data.normal_(0.0, std)

            # Normal Initialization for Biases
            layer.bias.data.normal_(0.0, 0.001)


class Discriminator(nn.Module):
    def __init__(self, input_dim):
        super(Discriminator, self).__init__()
        self.main = nn.Sequential(
            nn.Linear(input_dim, 512),
            nn.ReLU(True),
            nn.Linear(512, 128),
            nn.ReLU(True),
            nn.Linear(128, 1),
            nn.Sigmoid(),
        )

    def forward(self, input):
        return self.main(input)



from ..base import BaseModel
from abc import abstractmethod


class BaseAttacker(BaseModel):
    @classmethod
    def from_config(cls, name, arg_string, user_args, user_config):
        return super().from_config("attacker", name, arg_string, user_args, user_config)

    @abstractmethod
    def generate_fake(self, **kwargs):
        raise NotImplementedError



import torch
import torch.nn as nn
import torch.nn.functional as F
from .base import BaseVictim
from ...utils import VarDim, pick_optim
from ...default import MODEL


class MF(BaseVictim):
    def __init__(self, factor_num, embedding_size, dropout, **config):
        super(MF, self).__init__()
        num_users = config['dataset'].info_describe()['n_users']
        num_items = config['dataset'].info_describe()['n_items']
        self.dataset = config['dataset']
        self.config = config
        self.user_emb = nn.Embedding(num_users, embedding_size)
        self.user_bias = nn.Embedding(num_users, 1)
        self.item_emb = nn.Embedding(num_items, embedding_size)
        self.item_bias = nn.Embedding(num_items, 1)

        self.user_emb.weight.data.uniform_(0, 0.005)
        self.user_bias.weight.data.uniform_(-0.01, 0.01)
        self.item_emb.weight.data.uniform_(0, 0.005)
        self.item_bias.weight.data.uniform_(-0.01, 0.01)

        self.mean = nn.Parameter(torch.FloatTensor([factor_num]), False)
        self.dropout = nn.Dropout(dropout)

        self.optimizer = pick_optim(self.config['optim'])(
            self.parameters(), lr=self.config['lr']
        )
        self.loss_func = nn.BCEWithLogitsLoss()

    @classmethod
    def from_config(cls, **kwargs):
        args = list(MODEL['victim']['mf'])
        user_args = "dataset"
        return super().from_config("mf", args, user_args, kwargs)

    def forward(self, users, items):
        u_id = users
        i_id = items
        U = self.user_emb(u_id)
        b_u = self.user_bias(u_id).squeeze()
        I = self.item_emb(i_id)
        b_i = self.item_bias(i_id).squeeze()
        return self.dropout((U * I).sum(1) + b_u + b_i + self.mean)

    def train_step(self, **config):
        optim = self.optimizer
        total_loss = 0.0
        self.train()
        pbar = config.get('progress_bar', None)
        for idx, dp in enumerate(self.dataset.generate_batch()):
            user = dp['users']
            item = dp['items']
            label = dp['labels']

            pred = self(user, item)
            loss = self.loss_func(pred, label.float())

            optim.zero_grad()
            loss.backward()
            optim.step()

            total_loss += loss.item()
            if pbar:
                pbar.set_description(f"loss: {total_loss/(idx + 1):.4f}")
        return (total_loss / (idx + 1),)

    def input_describe(self):
        return {
            "train_step": {
                "users": (torch.int64, (VarDim(comment="batch"))),
                "items": (torch.int64, (VarDim(comment="batch"))),
                "labels": (torch.int64, (VarDim(comment="batch"))),
            },
            "forward": {
                "users": (torch.int64, (VarDim(comment="batch"))),
                "items": (torch.int64, (VarDim(comment="batch"))),
            },
        }

    def output_describe(self):
        return {
            "train_step": {
                "loss": (float, []),
            },
            "forward": {
                "unnormalized_scores": (torch.float32, [VarDim(comment="batch")])
            },
        }



# TODO
from .base import BaseAttacker
import math
import torch
from torch import nn
import torch.nn.functional as F
from torch import optim
import numpy as np
from ...default import MODEL
from ...utils import pick_optim, VarDim


def aia_attack_loss(logits, labels):
    log_probs = torch.nn.functional.log_softmax(logits, dim=-1)
    loss = -log_probs * labels
    instance_data = labels.sum(1)
    instance_loss = loss.sum(1)
    # Avoid divide by zeros.
    res = instance_loss / (instance_data + 0.1)  # PSILON)
    return res


def partial_update(loss, optimizer, part=0):
    optimizer.zero_grad()
    grad_groups = torch.autograd.grad(
        loss, optimizer.param_groups[part]['params'], allow_unused=True
    )
    for para_, grad_ in zip(optimizer.param_groups[part]['params'], grad_groups):
        if para_.grad is None:
            para_.grad = grad_
        else:
            # print(grad_)
            para_.grad.data = grad_

    optimizer.step()


class AIA(BaseAttacker):
    def __init__(self, **config) -> None:
        super().__init__()

        self.selected_ids = config['selected_ids']
        self.attack_num = config['attack_num']
        self.filler_num = config['filler_num']
        self.dataset = config['dataset']
        self.n_users = config['dataset'].info_describe()['n_users']
        self.n_items = config['dataset'].info_describe()['n_items']
        self.device = config['device']
        self.train_array = config['dataset'].info_describe()['train_mat']
        self.surrogate = config["surrogate_model"]
        self.config = config
        self.build_network()

    def build_network(self):
        sampled_idx = np.random.choice(
            np.where(np.sum(self.train_array > 0, 1) >= self.filler_num)[0],
            self.attack_num,
        )
        templates = self.train_array[sampled_idx]
        for idx, template in enumerate(templates):
            fillers = np.where(template)[0]
            np.random.shuffle(fillers)
            for iid in fillers[self.filler_num :]:
                templates[idx][iid] = 0.0
        self.real_template = torch.tensor(templates).type(torch.float).to(self.device)
        self.netG = RecsysGenerator(self.device, self.real_template).to(self.device)

        self.G_optimizer = pick_optim(self.config['optim_g'])(
            self.parameters(), lr=self.config['lr_g']
        )

    @classmethod
    def from_config(cls, **kwargs):
        args = list(MODEL['attacker']['aia'])
        user_args = "dataset"
        return super().from_config("aia", args, user_args, kwargs)

    def input_describe(self):
        return {"train_step": {"target_id_list": (list, VarDim())}}

    def output_describe(self):
        return {
            "train_step": {
                "g_losses": (float, []),
            }
        }

    def train_step(self, **config):
        target_id_list = config['target_id_list']
        self.netG.train()
        G_loss = torch.tensor(0.0).to(self.device)
        fake_tensor = self.netG(self.real_template)
        sur_predictions = self.get_sur_predictions(fake_tensor)
        for target_id in target_id_list:
            target_users = np.where(self.train_array[:, target_id] == 0)[0]
            attack_target = np.zeros((len(target_users), self.n_items))
            attack_target[:, target_id] = 1.0
            attack_target = (
                torch.from_numpy(attack_target).type(torch.float32).to(self.device)
            )
            higher_mask = (
                sur_predictions[target_users]
                >= (sur_predictions[target_users, target_id].reshape([-1, 1]))
            ).float()
            G_loss_sub = aia_attack_loss(
                logits=sur_predictions[target_users] * higher_mask,
                labels=attack_target,
            ).mean()
            G_loss += G_loss_sub
        G_loss = G_loss / 10

        partial_update(G_loss, self.G_optimizer)

        return (G_loss.cpu().item(),)

    def generate_fake(self, **config):
        target_id_list = config['target_id_list']
        with torch.no_grad():
            fake_tensor = self.netG(self.real_template)
            rate = int(self.attack_num / len(target_id_list))
            for i in range(len(target_id_list)):
                fake_tensor[i * rate : (i + 1) * rate, target_id_list[i]] = 5
        return fake_tensor.detach().cpu().numpy()

    def get_sur_predictions(self, fake_tensor):
        data_tensor = torch.cat(
            [
                torch.from_numpy(self.train_array).type(torch.float32).to(self.device),
                fake_tensor,
            ],
            dim=0,
        )

        surrogate = self.surrogate

        if surrogate == 'WMF':
            sur_trainer_ = WMFTrainer(
                n_users=self.n_users + self.attack_num,
                n_items=self.n_items,
                hidden_dim=self.config['hidden_dim_s'],
                # device=self.device,
                device=self.device,
                lr=self.config['lr_s'],
                weight_decay=self.config['weight_decay_s'],
                batch_size=self.config['batch_size_s'],
                weight_pos=self.config['weight_pos_s'],
                weight_neg=self.config['weight_neg_s'],
                verbose=False,
            )
            epoch_num_ = self.config['epoch_s']
            unroll_steps_ = self.config['unroll_steps_s']
        # elif surrogate == 'ItemAE':
        #     sur_trainer_ = ItemAETrainer(
        #         n_users=self.n_users + self.attack_num,
        #         n_items=self.n_items,
        #         hidden_dims=self.config.hidden_dim_s,
        #         device=self.device,
        #         lr=self.config.lr_s,
        #         l2=self.config.weight_decay_s,
        #         batch_size=self.config.batch_size_s,
        #         weight_pos=self.config.weight_pos_s,
        #         weight_neg=self.config.weight_neg_s,
        #         verbose=False)
        #     epoch_num_ = self.config.epoch_s
        #     unroll_steps_ = self.config.unroll_steps_s
        # elif surrogate == 'SVDpp':
        #     sur_trainer_ = SVDppTrainer(
        #         n_users=self.n_users + self.attack_num,
        #         n_items=self.n_items,
        #         hidden_dims=[128],
        #         device=self.device,
        #         lr=1e-3,
        #         l2=5e-2,
        #         batch_size=128,
        #         weight_alpha=20)
        #     epoch_num_ = 10
        #     unroll_steps_ = 1
        # elif surrogate == 'NMF':
        #     sur_trainer_ = NMFTrainer(
        #         n_users=self.n_users + self.attack_num,
        #         n_items=self.n_items,
        #         batch_size=128,
        #         device=self.device,
        #     )
        #     epoch_num_ = 50
        #     unroll_steps_ = 1
        # elif surrogate == 'PMF':
        #     sur_trainer_ = PMFTrainer(
        #         n_users=self.n_users + self.attack_num,
        #         n_items=self.n_items,
        #         hidden_dim=128,
        #         device=self.device,
        #         lr=0.0001,
        #         weight_decay=0.1,
        #         batch_size=self.config.batch_size_s,
        #         momentum=0.9,
        #         verbose=True)
        #     epoch_num_ = 50
        #     unroll_steps_ = 1
        else:
            raise ValueError(
                f'surrogate model error : {surrogate}',
            )

        sur_predictions = sur_trainer_.fit_adv(
            data_tensor=data_tensor, epoch_num=epoch_num_, unroll_steps=unroll_steps_
        )

        # sur_test_rmse = np.mean(
        #     (
        #         sur_predictions[: self.n_users][self.test_array > 0]
        #         .detach()
        #         .cpu()
        #         .numpy()
        #         - self.test_array[self.test_array > 0]
        #     )
        #     ** 2
        # )

        return sur_predictions


class WeightedMF(nn.Module):
    def __init__(self, n_users, n_items, hidden_dim):
        super(WeightedMF, self).__init__()

        self.n_users = n_users
        self.n_items = n_items
        self.dim = hidden_dim

        self.Q = nn.Parameter(
            torch.zeros([self.n_items, self.dim]).normal_(mean=0, std=0.1)
        )
        self.P = nn.Parameter(
            torch.zeros([self.n_users, self.dim]).normal_(mean=0, std=0.1)
        )
        self.params = nn.ParameterList([self.Q, self.P])

    def forward(self, user_id=None, item_id=None):
        if user_id is None and item_id is None:
            return torch.mm(self.P, self.Q.t())
        if user_id is not None:
            return torch.mm(self.P[[user_id]], self.Q.t())
        if item_id is not None:
            return torch.mm(self.P, self.Q[[item_id]].t())


class BaseTrainer(object):
    def __init__(self):
        self.args = None

        self.n_users = None
        self.n_items = None

        self.net = None
        self.optimizer = None
        self.metrics = None
        self.golden_metric = "Recall@50"

    @staticmethod
    def minibatch(*tensors, **kwargs):
        """Mini-batch generator for pytorch tensor."""
        batch_size = kwargs.get('batch_size', 128)

        if len(tensors) == 1:
            tensor = tensors[0]
            for i in range(0, len(tensor), batch_size):
                yield tensor[i : i + batch_size]
        else:
            for i in range(0, len(tensors[0]), batch_size):
                yield tuple(x[i : i + batch_size] for x in tensors)

    @staticmethod
    def mult_ce_loss(data, logits):
        """Multi-class cross-entropy loss."""
        log_probs = F.log_softmax(logits, dim=-1)
        loss = -log_probs * data

        instance_data = data.sum(1)
        instance_loss = loss.sum(1)
        # Avoid divide by zeros.
        res = instance_loss / (instance_data + 0.1)  # PSILON)
        return res

    @staticmethod
    def weighted_mse_loss(data, logits, weight_pos=1, weight_neg=0):
        """Mean square error loss."""
        weights = torch.ones_like(data) * weight_neg
        weights[data > 0] = weight_pos
        res = weights * (data - logits) ** 2
        return res.sum(1)

    @staticmethod
    def _array2sparsediag(x):
        values = x
        indices = np.vstack([np.arange(x.size), np.arange(x.size)])

        i = torch.LongTensor(indices)
        v = torch.FloatTensor(values)
        shape = [x.size, x.size]

        return torch.sparse.FloatTensor(i, v, torch.Size(shape))

    @property
    def _initialized(self):
        return self.net is not None

    def _initialize(self):
        """Initialize model and optimizer."""
        # See actual implementation in each trainer.
        raise NotImplementedError

    def recommend(self, data, top_k, return_preds=False, allow_repeat=False):
        """Generate a top-k recommendation (ranked) list."""
        # See actual implementation in each trainer.
        raise NotImplementedError

    def train_epoch(self, data):
        """Train model for one epoch"""
        # See actual implementation in each trainer.
        raise NotImplementedError

    def train_epoch_wrapper(self, train_data, epoch_num):
        """Wrapper for train_epoch with some logs."""
        epoch_loss = self.train_epoch(train_data)

    def evaluate_epoch(self, train_data, test_data, epoch_num):
        """Evaluate model performance on test data."""

        n_rows = train_data.shape[0]
        n_evaluate_users = test_data.shape[0]

        total_metrics_len = sum(len(x) for x in self.metrics)
        total_val_metrics = np.zeros([n_rows, total_metrics_len], dtype=np.float32)

        recommendations = self.recommend(train_data, top_k=100)

        valid_rows = list()
        for i in range(train_data.shape[0]):
            # Ignore augmented users, evaluate only on real users.
            if i >= n_evaluate_users:
                continue
            targets = test_data[i].indices
            if targets.size <= 0:
                continue

            recs = recommendations[i].tolist()

            metric_results = list()
            for metric in self.metrics:
                result = metric(targets, recs)
                metric_results.append(result)
            total_val_metrics[i, :] = np.concatenate(metric_results)
            valid_rows.append(i)

        # Average evaluation results by user.
        total_val_metrics = total_val_metrics[valid_rows]
        avg_val_metrics = (total_val_metrics.mean(axis=0)).tolist()

    def fit(self, train_data, test_data):
        """Full model training loop."""
        if not self._initialized:
            self._initialize()

        if self.args.save_feq > self.args.epochs:
            raise ValueError(
                "Model save frequency should be smaller than" " total training epochs."
            )

        start_epoch = 1
        best_checkpoint_path = ""
        best_perf = 0.0
        for epoch_num in range(start_epoch, self.args.epochs + 1):
            # Train the model.
            self.train_epoch_wrapper(train_data, epoch_num)
            if epoch_num % self.args.save_feq == 0:
                result = self.evaluate_epoch(train_data, test_data, epoch_num)

        # Load best model and evaluate on test data.
        self.restore(best_checkpoint_path)
        self.evaluate_epoch(train_data, test_data, -1)
        return

    def restore(self, path):
        return


class WMFTrainer(BaseTrainer):
    def __init__(
        self,
        n_users,
        n_items,
        device,
        hidden_dim,
        lr,
        weight_decay,
        batch_size,
        weight_pos,
        weight_neg,
        verbose=False,
    ):
        super(WMFTrainer, self).__init__()
        self.device = device
        #
        self.n_users = n_users
        self.n_items = n_items
        #
        self.hidden_dim = hidden_dim
        #
        self.lr = lr
        self.weight_decay = weight_decay
        self.batch_size = batch_size
        #
        self.weight_pos = weight_pos
        self.weight_neg = weight_neg
        #
        self.verbose = verbose

        pass

    def _initialize(self):
        self.net = WeightedMF(
            n_users=self.n_users, n_items=self.n_items, hidden_dim=self.hidden_dim
        ).to(self.device)

        self.optimizer = optim.Adam(
            self.net.parameters(), lr=self.lr, weight_decay=self.weight_decay
        )
        self.dim = self.net.dim

    def fit_adv(self, data_tensor, epoch_num, unroll_steps):
        self._initialize()

        import higher

        if not data_tensor.requires_grad:
            raise ValueError(
                "To compute adversarial gradients, data_tensor "
                "should have requires_grad=True."
            )
        #
        data_tensor = data_tensor.to(self.device)
        n_rows = data_tensor.shape[0]
        idx_list = np.arange(n_rows)
        #
        model = self.net.to(self.device)
        #
        for i in range(1, epoch_num - unroll_steps + 1):
            np.random.shuffle(idx_list)
            model.train()
            epoch_loss = 0.0

            for batch_idx in self.minibatch(idx_list, batch_size=self.batch_size):
                # Compute loss

                # TODO detach()
                loss = self.weighted_mse_loss(
                    data=data_tensor[batch_idx].detach(),
                    logits=model(user_id=batch_idx),
                    weight_pos=self.weight_pos,
                    weight_neg=self.weight_neg,
                ).sum()
                epoch_loss += loss.item()
                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()

        with higher.innerloop_ctx(model, self.optimizer) as (fmodel, diffopt):
            for i in range(epoch_num - unroll_steps + 1, epoch_num + 1):
                np.random.shuffle(idx_list)
                fmodel.train()
                epoch_loss = 0.0
                for batch_idx in self.minibatch(idx_list, batch_size=self.batch_size):
                    # Compute loss
                    # ===========warning=================

                    loss = self.weighted_mse_loss(
                        data=data_tensor[batch_idx],
                        logits=fmodel(user_id=batch_idx),
                        weight_pos=self.weight_pos,
                        weight_neg=self.weight_neg,
                    ).sum()
                    # ====================================
                    epoch_loss += loss.item()
                    diffopt.step(loss)

            fmodel.eval()
            predictions = fmodel()
        return predictions  # adv_loss  # .item(), adv_grads[-n_fakes:, ]

    def recommend(self, data, top_k, return_preds=False, allow_repeat=False):
        # Set model to eval mode
        model = self.net.to(self.device)
        model.eval()

        n_rows = data.shape[0]
        idx_list = np.arange(n_rows)
        recommendations = np.empty([n_rows, top_k], dtype=np.int64)
        all_preds = list()
        with torch.no_grad():
            for batch_idx in self.minibatch(
                idx_list, batch_size=self.args.valid_batch_size
            ):
                batch_data = data[batch_idx].toarray()

                preds = model(user_id=batch_idx)
                if return_preds:
                    all_preds.append(preds)
                if not allow_repeat:
                    preds[batch_data.nonzero()] = -np.inf
                if top_k > 0:
                    _, recs = preds.topk(k=top_k, dim=1)
                    recommendations[batch_idx] = recs.cpu().numpy()

        if return_preds:
            return recommendations, torch.cat(all_preds, dim=0).cpu()
        else:
            return recommendations


class BaseGenerator(nn.Module):
    def __init__(self, device, input_dim):
        super(BaseGenerator, self).__init__()
        #
        self.input_dim = input_dim
        self.device = device

        """helper_tensor"""

        self.epsilon = torch.tensor(1e-4).to(self.device)  # 计算boundary
        self.helper_tensor = torch.tensor(2.5).to(device)
        pass

    def project(self, fake_tensor):
        fake_tensor.data = torch.round(fake_tensor)
        # fake_tensor.data = torch.where(fake_tensor < 1, torch.ones_like(fake_tensor).to(self.device), fake_tensor)
        fake_tensor.data = torch.where(
            fake_tensor < 0, torch.zeros_like(fake_tensor).to(self.device), fake_tensor
        )
        fake_tensor.data = torch.where(
            fake_tensor > 5, torch.tensor(5.0).to(self.device), fake_tensor
        )
        #
        return fake_tensor

    def forward(self, input):
        raise NotImplementedError


class RecsysGenerator(BaseGenerator):
    def __init__(self, device, init_tensor):
        super(RecsysGenerator, self).__init__(device, init_tensor.shape[1])
        """
        fake_parameter
        """
        fake_tensor = init_tensor.clone().detach().requires_grad_(True)
        self.fake_parameter = torch.nn.Parameter(fake_tensor, requires_grad=True)
        self.register_parameter("fake_tensor", self.fake_parameter)
        pass

    def forward(self, input=None):
        return self.project(self.fake_parameter * (input > 0))
